# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import warnings

# bitsandbytes is a very noisy library. A lot of it is print statements that we can't easily suppress,
# but we can at least suppress a bunch of spurious warnings.
warnings.filterwarnings('ignore', category=UserWarning, module='bitsandbytes')

import logging

try:
    from flash_attn import flash_attn_func
    del flash_attn_func
except ImportError as e:
    if 'undefined symbol' in str(e):
        raise ImportError(
            'The flash_attn package is not installed correctly. Usually this means that your runtime version'
            +
            ' of PyTorch is different from the version that flash_attn was installed with, which can occur when your'
            +
            ' workflow has resulted in PyTorch being reinstalled. This probably happened because you are using an old Docker image'
            +
            ' with the latest version of LLM Foundry. Check that the PyTorch version in your Docker image matches the PyTorch version'
            +
            ' in LLM Foundry setup.py and update accordingly. The latest Docker image can be found in the README.',
        ) from e

from llmfoundry.utils.logging_utils import SpecificWarningFilter

# Filter out Hugging Face warning for not using a pinned revision of the model
logger = logging.getLogger('transformers.dynamic_module_utils')
new_files_warning_filter = SpecificWarningFilter(
    'A new version of the following files was downloaded from',
)

logger.addFilter(new_files_warning_filter)

from llmfoundry import (
    algorithms,
    callbacks,
    cli,
    data,
    eval,
    interfaces,
    loggers,
    metrics,
    models,
    optim,
    tokenizers,
    tp,
    utils,
)
from llmfoundry._version import __version__
from llmfoundry.data import StreamingFinetuningDataset, StreamingTextDataset
from llmfoundry.eval import InContextLearningDataset, InContextLearningMetric
from llmfoundry.models.hf import ComposerHFCausalLM
from llmfoundry.models.mpt import (
    ComposerMPTCausalLM,
    MPTConfig,
    MPTForCausalLM,
    MPTModel,
    MPTPreTrainedModel,
)
from llmfoundry.optim import DecoupledLionW

__all__ = [
    '__version__',
    'StreamingFinetuningDataset',
    'StreamingTextDataset',
    'InContextLearningDataset',
    'InContextLearningMetric',
    'ComposerHFCausalLM',
    'MPTConfig',
    'MPTPreTrainedModel',
    'MPTModel',
    'MPTForCausalLM',
    'ComposerMPTCausalLM',
    'DecoupledLionW',
    'algorithms',
    'callbacks',
    'cli',
    'data',
    'eval',
    'interfaces',
    'loggers',
    'metrics',
    'models',
    'optim',
    'tokenizers',
    'tp',
    'utils',
]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Optional, Union

import torch

from llmfoundry.layers_registry import (
    attention_classes,
    fcs,
    ffns,
    ffns_with_megablocks,
    ffns_with_norm,
    norms,
)
from llmfoundry.utils.registry_utils import construct_from_registry

__all__ = [
    'build_attention_layer',
    'build_ffn',
    'build_fc',
    'build_norm',
]


def build_norm(
    name: str,
    normalized_shape: Union[int, list[int], torch.Size],
    eps: Optional[float] = 1e-5,
    device: Optional[str] = None,
):
    kwargs = {
        'normalized_shape': normalized_shape,
        'eps': eps,
        'device': device,
    }

    return construct_from_registry(
        name=name,
        registry=norms,
        pre_validation_function=torch.nn.Module,
        kwargs=kwargs,
    )


def build_ffn(
    name: str,
    d_model: int,
    expansion_ratio: float,
    device: Optional[str],
    bias: bool,
    ffn_kwargs: dict[str, Any],
):

    registry_to_use = ffns
    if name in ffns_with_norm:
        registry_to_use = ffns_with_norm

    if name in ffns_with_megablocks:
        registry_to_use = ffns_with_megablocks

    kwargs = {
        'd_model': d_model,
        'expansion_ratio': expansion_ratio,
        'device': device,
        'bias': bias,
        **{k: v for k, v in ffn_kwargs.items() if k != 'ffn_type'},
    }

    def _validation_function(maybe_module: Any):
        if not isinstance(maybe_module, torch.nn.Module):
            raise ValueError(f'Function {name} must return a torch.nn.Module.')

    result = construct_from_registry(
        name=name,
        registry=registry_to_use,
        post_validation_function=_validation_function,
        partial_function=False,
        kwargs=kwargs,
    )

    if name in ffns_with_norm:
        result._has_norm = True

    if name in ffns_with_megablocks:
        result._uses_megablocks = True

    return result


def build_attention_layer(
    name: str,
    attn_kwargs: dict[str, Any],
):
    return construct_from_registry(
        name=name,
        registry=attention_classes,
        pre_validation_function=torch.nn.Module,
        kwargs=attn_kwargs,
    )


def build_fc(
    name: str,
    in_features: int,
    out_features: int,
    fc_kwargs: dict[str, Any],
):
    kwargs = {
        'in_features': in_features,
        'out_features': out_features,
        **{k: v for k, v in fc_kwargs.items() if k != 'name'},
    }

    return construct_from_registry(
        name=name,
        registry=fcs,
        pre_validation_function=torch.nn.Module,
        kwargs=kwargs,
    )



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import contextlib
import copy
import logging
import math
import os
import re
import shutil
import tempfile
import time
import warnings
from multiprocessing.context import SpawnProcess
from pathlib import Path
from typing import Any, Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn
from composer.core import Callback, Event, Precision, State, Time, TimeUnit
from composer.devices import Device
from composer.loggers import Logger, MLFlowLogger
from composer.models import HuggingFaceModel
from composer.utils import (
    dist,
    format_name_with_dist_and_time,
    maybe_create_remote_uploader_downloader_from_uri,
    parse_uri,
)
from composer.utils.misc import create_interval_scheduler
from mlflow.transformers import _fetch_model_card, _write_license_information
from torch.distributed._tensor import DTensor
from torch.distributed.checkpoint.state_dict import (
    StateDictOptions,
    get_model_state_dict,
)
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from transformers import (
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizerBase,
)

from llmfoundry.models.mpt import MPTConfig, MPTForCausalLM
from llmfoundry.models.utils import init_empty_weights
from llmfoundry.utils.exceptions import StoragePermissionError
from llmfoundry.utils.huggingface_hub_utils import \
    edit_files_for_hf_compatibility

try:
    import transformer_engine.pytorch as te
    is_te_imported = True
except ModuleNotFoundError:
    is_te_imported = False

log = logging.getLogger(__name__)

__all__ = ['HuggingFaceCheckpointer']

_LICENSE_FILE_PATTERN = re.compile(r'license(\.[a-z]+|$)', re.IGNORECASE)

from contextlib import contextmanager


@contextmanager
def _monitor_process_saver(mlflow_logger: MLFlowLogger):
    # Save the current monitor process
    if hasattr(mlflow_logger, 'monitor_process'):
        original_monitor_process = mlflow_logger.monitor_process  # type: ignore
        mlflow_logger.monitor_process = None  # type: ignore
    else:
        original_monitor_process = None

    try:
        # Yield control back to the calling code
        yield
    finally:
        # Restore the monitor process
        if original_monitor_process is not None:
            mlflow_logger.monitor_process = original_monitor_process  # type: ignore


def _maybe_get_license_filename(
    local_dir: str,
    pretrained_model_name: Optional[str] = None,
) -> Optional[str]:
    """Returns the name of the license file if it exists in the local_dir.

    Note: This is intended to be consistent with the code in MLflow.
    https://github.com/mlflow/mlflow/blob/5d13d6ec620a02de9a5e31201bf1becdb9722ea5/mlflow/transformers/__init__.py#L1152

    Since LLM Foundry supports local model files being used rather than fetching the files from the Hugging Face Hub,
    MLflow's logic to fetch and write the license information on model save is not applicable; it will try to search for
    a Hugging Face repo named after the local path. However, the user can provide the original pretrained model name,
    in which case this function will use that to fetch the correct license information.

    If the license file does not exist, returns None.
    """
    try:
        license_filename = next(
            file for file in os.listdir(local_dir)
            if _LICENSE_FILE_PATTERN.search(file)
        )

        # If a pretrained model name is provided, replace the license file with the correct info from HF Hub.
        if pretrained_model_name is not None:
            log.info(
                f'Overwriting license file {license_filename} with license info for model {pretrained_model_name} from Hugging Face Hub',
            )
            os.remove(os.path.join(local_dir, license_filename))
            model_card = _fetch_model_card(pretrained_model_name)

            local_dir_path = Path(local_dir).absolute()
            _write_license_information(
                pretrained_model_name,
                model_card,
                local_dir_path,
            )
            license_filename = next(
                file for file in os.listdir(local_dir)
                if _LICENSE_FILE_PATTERN.search(file)
            )

        return license_filename

    except StopIteration:
        return None


def _log_model_with_multi_process(
    mlflow_logger: MLFlowLogger,
    python_logging_level: int,
    transformers_model: str,
    artifact_path: str,
    pretrained_model_name: str,
    registered_model_name: Optional[str],
    await_registration_for: int,
    mlflow_logging_config: dict[str, Any],
):
    """Call MLFlowLogger.log_model.

    First, patch the mlflow save_model function by removing duplicate tokenizer
    files in the model directory. Then, register the model to mlflow from a
    child process.
    """
    # Setup logging for child process. This ensures that any logs from composer are surfaced.
    if python_logging_level > 0:
        # If logging_level is 0, then the composer logger was unset.
        logging.basicConfig(
            format=
            f'%(asctime)s: rank{dist.get_global_rank()}[%(process)d][%(threadName)s]: %(levelname)s: %(name)s: %(message)s',
            force=True,
        )
        logging.getLogger('composer').setLevel(python_logging_level)
        logging.getLogger('llmfoundry').setLevel(python_logging_level)

    import mlflow
    original_save_model = mlflow.transformers.save_model

    def save_model_patch(*args: Any, **kwargs: Any):
        original_save_model(*args, **kwargs)
        tokenizer_files = []
        save_path = kwargs['path']
        tokenizer_path = os.path.join(save_path, 'components', 'tokenizer')
        if os.path.exists(tokenizer_path):
            tokenizer_files = os.listdir(
                os.path.join(save_path, 'components', 'tokenizer'),
            )
        try:
            # Check if there are duplicate tokenizer files in the model directory and remove them.
            for tokenizer_file_name in tokenizer_files:
                dupe_file = os.path.isfile(
                    os.path.join(save_path, 'model', tokenizer_file_name),
                )
                if dupe_file:
                    log.debug(
                        f'Removing duplicate tokenizer file: {tokenizer_file_name}',
                    )
                    os.remove(
                        os.path.join(save_path, 'model', tokenizer_file_name),
                    )
            license_filename = _maybe_get_license_filename(
                save_path,
                pretrained_model_name,
            )
            if license_filename is not None:
                mlflow_logger._mlflow_client.log_artifact(
                    mlflow_logger._run_id,
                    os.path.join(save_path, license_filename),
                )
        except Exception as e:
            log.error(
                f'Exception when removing duplicate tokenizer files in the model directory',
                e,
            )

    mlflow.transformers.save_model = save_model_patch  # type: ignore

    mlflow.set_tracking_uri(mlflow_logger.tracking_uri)
    if mlflow_logger.model_registry_uri is not None:
        mlflow.set_registry_uri(mlflow_logger.model_registry_uri)

    register_model_path = f'{mlflow_logger.model_registry_prefix}.{registered_model_name}' if mlflow_logger.model_registry_prefix and registered_model_name else registered_model_name
    mlflow_logger.log_model(
        transformers_model=transformers_model,
        flavor='transformers',
        artifact_path=artifact_path,
        registered_model_name=register_model_path,
        run_id=mlflow_logger._run_id,
        await_registration_for=await_registration_for,
        **mlflow_logging_config,
    )


def _register_model_with_run_id_multiprocess(
    mlflow_logger: MLFlowLogger,
    composer_logging_level: int,
    model_uri: str,
    name: str,
    await_creation_for: int,
):
    """Call MLFlowLogger.register_model_with_run_id.

    Used mainly to register from a child process.
    """
    # Setup logging for child process. This ensures that any logs from composer are surfaced.
    if composer_logging_level > 0:
        # If logging_level is 0, then the composer logger was unset.
        logging.basicConfig(
            format=
            f'%(asctime)s: rank{dist.get_global_rank()}[%(process)d][%(threadName)s]: %(levelname)s: %(name)s: %(message)s',
            force=True,
        )
        logging.getLogger('composer').setLevel(composer_logging_level)

    # Register model.
    mlflow_logger.register_model_with_run_id(
        model_uri=model_uri,
        name=name,
        await_creation_for=await_creation_for,
    )


class HuggingFaceCheckpointer(Callback):
    """Save a huggingface formatted checkpoint during training.

    Args:
        save_folder (str): Top level folder to save checkpoints to (can be a
            URI). It is likely that this would be the same as your save_folder.
        save_interval: Union[str, int, Time]: The interval describing how often
            checkpoints should be saved. If an integer, it will be assumed to be
            in :attr:`.TimeUnit.EPOCH`. Otherwise, the unit must be either
            :attr:`.TimeUnit.EPOCH`, :attr:`.TimeUnit.BATCH`,
            :attr:`.TimeUnit.TOKEN`, or :attr:`.TimeUnit.SAMPLE`.
        huggingface_folder_name (str): Folder to save each checkpoint under (can
            be a format string). Default is ``ba{batch}``.
        precision: The precision to save the model in. Default is ``float32``.
            Options are ``bfloat16``, ``float16``, or ``float32``.
        overwrite (bool): Whether to overwrite previous checkpoints.
        mlflow_registered_model_name (Optional[str]): The name to register the
            model under in the MLflow model registry. If ``None``, the model
            will not be registered. Default is ``None``.
        mlflow_logging_config (Optional[dict]): A dictionary of config arguments
            that will get passed along to the MLflow ``save_model`` call.
            Expected to contain ``metadata`` and ``task`` keys. If either is
            unspecified, the defaults are ``'text-generation'`` and
            ``{'task': 'llm/v1/completions'}`` respectively. A default input example
            and signature intended for text generation is also included under the
            keys ``input_example`` and ``signature``.
        flatten_imports (Sequence[str]): A sequence of import prefixes that will
            be flattened when editing MPT files.
        final_register_only (bool): If true, only register the model in the MLFlow
            registry on the last batch and do not save the HuggingFace checkpoint. If
            registration fails or mlflow_registered_model_name is not set, then we will
            fallback to saving the HuggingFace checkpoint.
    """

    def __init__(
        self,
        save_folder: str,
        save_interval: Union[str, int, Time],
        huggingface_folder_name: str = 'ba{batch}',
        precision: str = 'float32',
        overwrite: bool = True,
        mlflow_registered_model_name: Optional[str] = None,
        mlflow_logging_config: Optional[dict] = None,
        flatten_imports: Sequence[str] = ('llmfoundry',),
        final_register_only: bool = False,
        register_wait_seconds: int = 7200,
    ):
        _, _, self.save_dir_format_str = parse_uri(save_folder)
        self.overwrite = overwrite
        self.precision = precision
        self.dtype = {
            'float32': torch.float32,
            'float16': torch.float16,
            'bfloat16': torch.bfloat16,
        }[precision]
        self.flatten_imports = flatten_imports
        self.using_peft = False

        self.final_register_only = final_register_only
        self.register_wait_seconds = register_wait_seconds

        self.mlflow_registered_model_name = mlflow_registered_model_name
        if self.final_register_only and self.mlflow_registered_model_name is None:
            self.final_register_only = False
            warnings.warn(
                'final_register_only is set to True, but mlflow_registered_model_name is not set. '
                +
                f'Defaulting to final_register_only=False and saving the HuggingFace checkpoint to {save_folder=}.',
            )

        # mlflow config setup
        if mlflow_logging_config is None:
            mlflow_logging_config = {}
        if self.mlflow_registered_model_name is not None:
            # Both the metadata and the task are needed in order for mlflow
            # and databricks optimized model serving to work
            passed_metadata = mlflow_logging_config.get('metadata', {})
            mlflow_logging_config['metadata'] = passed_metadata
            mlflow_logging_config.setdefault('task', 'llm/v1/completions')

            default_input_example = {
                'prompt': np.array(['What is Machine Learning?']),
            }
            is_chat = mlflow_logging_config['task'].endswith('chat') or (
                mlflow_logging_config['metadata'] is not None and
                mlflow_logging_config['metadata'].get('task',
                                                      '').endswith('chat')
            )
            if is_chat:
                default_input_example = {
                    'messages': [{
                        'role': 'user',
                        'content': 'What is Machine Learning?',
                    }],
                }
            mlflow_logging_config.setdefault(
                'input_example',
                default_input_example,
            )

        self.mlflow_logging_config = mlflow_logging_config
        if 'metadata' in self.mlflow_logging_config:
            self.pretrained_model_name = self.mlflow_logging_config[
                'metadata'].get(
                    'pretrained_model_name',
                    None,
                )
        else:
            self.pretrained_model_name = None

        self.huggingface_folder_name_fstr = os.path.join(
            'huggingface',
            huggingface_folder_name,
        )

        self.save_interval: Time = Time.from_input(
            save_interval,
            TimeUnit.EPOCH,
        )
        self.check_interval = create_interval_scheduler(
            self.save_interval,
            include_end_of_training=True,
        )
        self.remote_ud = maybe_create_remote_uploader_downloader_from_uri(
            save_folder,
            loggers=[],
        )
        if self.remote_ud is not None:
            self.remote_ud._num_concurrent_uploads = 4

        self.last_checkpoint_batch: Optional[Time] = None
        self.mlflow_loggers = []

        self.register_processes: list[SpawnProcess] = []
        # Temporary save directory used by child_processes.
        self.temp_save_dir = None

    def run_event(self, event: Event, state: State, logger: Logger) -> None:
        # The interval scheduler handles only returning True for the appropriate events
        if state.get_elapsed_duration() is not None and self.check_interval(
            state,
            event,
        ) and self.last_checkpoint_batch != state.timestamp.batch:
            is_last_batch = self._is_last_batch(state)
            self._save_checkpoint(
                state,
                logger,
                register_to_mlflow=(
                    self.mlflow_registered_model_name is not None and
                    is_last_batch
                ),
                upload_to_save_folder=not self.final_register_only or
                not is_last_batch,
            )
        elif event == Event.INIT:
            if not isinstance(state.model, HuggingFaceModel):
                raise ValueError(
                    f'`HuggingFaceCheckpointer` is only compatible with `HuggingFaceModel`s. '
                    + f'Got {type(state.model)} instead.',
                )
            if self.remote_ud is not None:
                try:
                    self.remote_ud.init(state, logger)
                except PermissionError as e:
                    if 'Client Error' in str(
                        e,
                    ):  # thrown from composer.utils._wrap_mlflow_exceptions
                        raise StoragePermissionError(
                            'Error when write to save_folder.',
                        ) from e
                    raise e
                state.callbacks.append(self.remote_ud)

            if self.mlflow_registered_model_name is not None:
                self.mlflow_loggers = [
                    logger_destination
                    for logger_destination in logger.destinations
                    if isinstance(logger_destination, MLFlowLogger)
                ]
                if len(self.mlflow_loggers) == 0:
                    raise ValueError(
                        f'`mlflow_registered_model_name` was set, but no `MLFlowLogger` was found in the `logger.destinations` list. '
                        +
                        'Please add an `MLFlowLogger` or set `mlflow_registered_model_name` to `None`.',
                    )

                import mlflow
                mlflow.environment_variables.MLFLOW_HUGGINGFACE_MODEL_MAX_SHARD_SIZE.set(
                    '1GB',
                )

            # Check if the model is using PEFT
            if state.is_model_ddp:
                composer_model = state.model.module
            elif isinstance(state.model.model, FSDP):
                composer_model = state.model
            else:
                composer_model = state.model
            self.using_peft = composer_model.using_peft
        elif event == Event.FIT_END:
            # Wait for all child processes spawned by the callback to finish.
            timeout = self.register_wait_seconds
            wait_start = time.time()
            while not self._all_register_processes_done(state.device):
                wait_time = time.time() - wait_start
                if wait_time > timeout:
                    raise TimeoutError(
                        f'Waited {wait_time} seconds for child processes to complete. Exceeded timeout of {timeout} seconds.',
                    )
                time.sleep(2)

            if self._any_register_processes_error(
                state.device,
            ) and self.final_register_only:
                log.error(
                    'An error occurred in one or more registration processes. Fallback to saving the HuggingFace checkpoint.',
                )
                self._save_checkpoint(
                    state,
                    logger,
                    upload_to_save_folder=True,
                    register_to_mlflow=False,
                )

            # Clean up temporary save directory; all processes are done with it.
            if self.temp_save_dir is not None:
                shutil.rmtree(self.temp_save_dir)

    def _is_last_batch(self, state: State):
        elapsed_duration = state.get_elapsed_duration()
        if elapsed_duration is not None and elapsed_duration >= 1.0:
            return True

        assert state.max_duration is not None  # for pyright

        epoch_complete = state.dataloader_len == state.timestamp.batch_in_epoch
        second_to_last_epoch = state.max_duration.unit == TimeUnit.EPOCH and (
            state.timestamp.epoch == state.max_duration.value - 1
        )
        # If the save interval is specified as exactly the same number of batches as the total duration,
        # but the max duration is specified in epochs, we need a special case to identify we are on the last batch
        # and should write the mlflow checkpoint. This should occur on the last batch of the final epoch.
        if self.save_interval.unit == TimeUnit.BATCH and second_to_last_epoch and epoch_complete:
            return True

        # If the save interval is specified as 1dur, and the max duration is in epoch units
        # we need a special case to identify we are on the last batch and should write the mlflow checkpoint
        if self.save_interval.unit == TimeUnit.DURATION and self.save_interval.value == 1 and state.max_duration.unit == TimeUnit.EPOCH:
            assert state.dataloader_len is not None  # for pyright
            return int(state.timestamp.batch) % math.ceil(
                state.max_duration.value * state.dataloader_len,
            ) == 0

        return False

    def _all_register_processes_done(self, device: Device) -> bool:
        not_done = any(
            process.is_alive() for process in self.register_processes
        )
        x = device.tensor_to_device(torch.tensor(1 if not_done else 0))
        dist.all_reduce(x, reduce_operation='MAX')
        return x.item() == 0

    def _any_register_processes_error(self, device: Device) -> bool:
        has_errors = any(
            process.exitcode is not None and process.exitcode != 0
            for process in self.register_processes
        )
        x = device.tensor_to_device(torch.tensor(1 if has_errors else 0))
        dist.all_reduce(x, reduce_operation='MAX')
        return x.item() == 1

    def transform_model_and_tokenizer(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
    ) -> tuple[PreTrainedModel, PreTrainedTokenizerBase]:
        """Transform the model and tokenizer before saving.

        This allows a subclass to modify the model and tokenizer before saving. The base class implementation will
        make no modifications.

        Args:
            model (PreTrainedModel): The model to be transformed.
            tokenizer (PreTrainedTokenizerBase): The tokenizer to be transformed.

        Returns:
            Tuple[PreTrainedModel, PreTrainedTokenizerBase]: The transformed model and tokenizer.
        """
        return model, tokenizer

    def transform_config(
        self,
        original_config: PretrainedConfig,
    ) -> PretrainedConfig:
        """Transform the model config before saving.

        Args:
            original_config (Any): The original model config.

        Returns:
            The transformed model config.
        """
        copied_config = copy.deepcopy(original_config)
        if copied_config.model_type == 'mpt':
            copied_config.attn_config['attn_impl'] = 'torch'
            copied_config.init_device = 'cpu'
            if 'moe_world_size' in getattr(copied_config, 'ffn_config', {}):
                copied_config.ffn_config['moe_world_size'] = 1
        return copied_config

    def pre_register_edit(self, local_save_path: str):
        """Edit the model before registering with MLflow.

        This allows a subclass to modify the model before registering with MLflow. The base class implementation will
        make no modifications.

        Args:
            local_save_path (str): The path to the model to be transformed.
        """
        pass

    def transform_model_pre_registration(
        self,
        model: PreTrainedModel,
    ) -> PreTrainedModel:
        """Transform the model before registering with MLflow.

        This allows a subclass to modify the model before registering with MLflow. The base class implementation will
        make no modifications.

        Args:
            model (PreTrainedModel): The model to be transformed.

        Returns:
            PreTrainedModel: The transformed model.
        """
        return model

    def _save_checkpoint(
        self,
        state: State,
        logger: Logger,
        upload_to_save_folder: bool,
        register_to_mlflow: bool,
    ):
        """Save a HuggingFace formatted checkpoint.

        Args:
            state (State): The training state.
            logger (Logger): The logger.
            upload_to_save_folder (bool): Whether to upload the HF checkpoint to the save folder.
            register_to_mlflow (bool): Whether to register the model to MLFlow
        """
        del logger  # unused

        self.last_checkpoint_batch = state.timestamp.batch

        log.info('Saving HuggingFace formatted checkpoint')

        from transformers.models.auto.configuration_auto import CONFIG_MAPPING
        CONFIG_MAPPING._extra_content['mpt'] = MPTConfig
        MPTConfig.register_for_auto_class()
        MPTForCausalLM.register_for_auto_class('AutoModelForCausalLM')

        save_dir = format_name_with_dist_and_time(
            str(
                Path(self.save_dir_format_str) /
                self.huggingface_folder_name_fstr,
            ),
            state.run_name,
            state.timestamp,
        )

        # Use a temporary directory if save_dir is remote.
        use_temp_dir = self.remote_ud is not None
        temp_save_dir = tempfile.mkdtemp() if use_temp_dir else save_dir

        log.debug('Gathering state dict')

        if state.is_model_ddp:
            original_model: PreTrainedModel = state.model.module.model
            state_dict_model = state.model.module.model
            original_tokenizer = state.model.module.tokenizer
        elif isinstance(state.model.model, FSDP):
            original_model: PreTrainedModel = state.model.model.module
            state_dict_model = state.model.model
            original_tokenizer = state.model.tokenizer
        else:
            original_model: PreTrainedModel = state.model.model
            state_dict_model = state.model.model
            original_tokenizer = state.model.tokenizer

        cpu_offload = True

        # Add hook to move tensors to cpu to avoid CUDA OOM
        def tensor_hook(
            module: nn.Module,
            state_dict: dict[str, Any],
            prefix: str,
            *args: Any,
        ) -> dict[str, Any]:
            dtensor_fqns = []
            for fqn in state_dict.keys():
                tensor = state_dict[fqn]
                if isinstance(tensor, DTensor):
                    dtensor_fqns.append(fqn)
                    tensor = tensor.full_tensor()  # type: ignore
                    if dist.get_global_rank() == 0:
                        # Offload any DTensors to CPU
                        if cpu_offload:
                            tensor = tensor.cpu()
                        state_dict[fqn] = tensor
                    else:
                        state_dict[fqn] = None

                if isinstance(state_dict[fqn], torch.Tensor):
                    state_dict[fqn] = state_dict[fqn].to(dtype=self.dtype)
                del tensor
            if dist.get_global_rank() != 0:
                state_dict = {}
            return state_dict

        hooks = []
        for _, module in state_dict_model.named_modules():
            hooks.append(module._register_state_dict_hook(tensor_hook),)

        state_dict = get_model_state_dict(
            state_dict_model,
            options=StateDictOptions(
                full_state_dict=True,
                cpu_offload=cpu_offload,
            ),
        )
        for hook in hooks:
            hook.remove()

        new_model_instance = None  # Need this for pyright because variable could be unbound

        if dist.get_global_rank() == 0:
            log.debug('Saving Hugging Face checkpoint in global rank 0')

            # Transform HF config before building 2nd model copy
            new_config = self.transform_config(
                original_config=original_model.config,
            )

            log.debug(f'Creating new model instance')

            # First create the model instance on meta device to avoid the
            # initialization cost.
            with init_empty_weights():
                if self.using_peft:
                    active_adapter = original_model.active_adapter
                    base_model = original_model.get_base_model()
                    new_base_model_instance = type(base_model)(new_config)

                    new_model_instance = type(original_model)(
                        new_base_model_instance,
                        original_model.peft_config[active_adapter],
                    )
                    del new_base_model_instance
                else:
                    new_model_instance = type(original_model)(new_config)
                    if new_model_instance.generation_config is not None:
                        new_model_instance.generation_config.update(
                            **original_model.generation_config.to_dict(),
                        )

            # Then load the state dict in with "assign" so that the state dict
            # is loaded properly even though the model is initially on meta device.
            new_model_instance.load_state_dict(state_dict, assign=True)
            del state_dict

            # Transform the model and tokenizer before saving
            new_model_instance, original_tokenizer = self.transform_model_and_tokenizer(
                new_model_instance,
                original_tokenizer,
            )

            # Ensure that the pretrained model name is correctly set on the saved HF checkpoint.
            if self.pretrained_model_name is not None:
                new_model_instance.name_or_path = self.pretrained_model_name
                if self.using_peft:
                    new_model_instance.base_model.name_or_path = self.pretrained_model_name
                    for k in new_model_instance.peft_config.keys():
                        new_model_instance.peft_config[
                            k
                        ].base_model_name_or_path = self.pretrained_model_name

            log.debug('Saving Hugging Face checkpoint to disk')

            if upload_to_save_folder:
                # This context manager casts the TE extra state in io.BytesIO format to tensor format
                # Needed for proper hf ckpt saving.
                context_manager = te.onnx_export(
                    True,
                ) if is_te_imported and state.precision == Precision.AMP_FP8 else contextlib.nullcontext(
                )
                with context_manager:
                    new_model_instance.save_pretrained(temp_save_dir)
                if original_tokenizer is not None:
                    assert isinstance(
                        original_tokenizer,
                        PreTrainedTokenizerBase,
                    )
                    original_tokenizer.save_pretrained(temp_save_dir)

                # Only need to edit files for MPT because it has custom code
                if new_model_instance.config.model_type == 'mpt':
                    log.debug('Editing MPT files for HuggingFace compatibility')
                    edit_files_for_hf_compatibility(
                        temp_save_dir,
                        self.flatten_imports,
                    )

                if self.remote_ud is not None:
                    for filename in os.listdir(temp_save_dir):
                        remote_file_name = os.path.join(save_dir, filename)
                        remote_file_uri = self.remote_ud.remote_backend.get_uri(
                            remote_file_name,
                        )
                        log.info(
                            f'Uploading HuggingFace formatted checkpoint to {remote_file_uri}',
                        )
                        self.remote_ud.upload_file(
                            state=state,
                            remote_file_name=remote_file_name,
                            file_path=Path(
                                os.path.join(temp_save_dir, filename),
                            ),
                            overwrite=self.overwrite,
                        )

        dist.barrier()

        if dist.get_global_rank() == 0:
            if register_to_mlflow:
                if self.using_peft:

                    # Save and register peft model to mlflow, this code path uses our older two step logic
                    self._save_and_register_peft_model(
                        state,
                        new_model_instance,
                        original_tokenizer,
                        temp_save_dir,
                    )
                else:
                    register_save_dir = os.path.join(
                        temp_save_dir,
                        'register_save',
                    )
                    assert new_model_instance is not None
                    new_model_instance = self.transform_model_pre_registration(
                        new_model_instance,
                    )
                    new_model_instance.save_pretrained(register_save_dir)
                    if original_tokenizer:
                        original_tokenizer.save_pretrained(register_save_dir)

                    self.pre_register_edit(register_save_dir)

                    for mlflow_logger in self.mlflow_loggers:
                        if self.mlflow_registered_model_name:
                            log.debug(
                                f'Registering model to UC at {mlflow_logger.model_registry_prefix}.{self.mlflow_registered_model_name}',
                            )

                        # Save the monitor process to be restored after registering the model.
                        with _monitor_process_saver(mlflow_logger):
                            process = SpawnProcess(
                                target=_log_model_with_multi_process,
                                kwargs={
                                    'mlflow_logger':
                                        mlflow_logger,
                                    'python_logging_level':
                                        logging.getLogger('llmfoundry').level,
                                    'transformers_model':
                                        register_save_dir,
                                    'artifact_path':
                                        'final_model_checkpoint',
                                    'pretrained_model_name':
                                        self.pretrained_model_name,
                                    'registered_model_name':
                                        self.mlflow_registered_model_name,
                                    'await_registration_for':
                                        3600,
                                    'mlflow_logging_config':
                                        self.mlflow_logging_config,
                                },
                            )

                            process.start()
                            self.register_processes.append(process)

                # Save the temporary directory to be cleaned up later.
                if use_temp_dir:
                    self.temp_save_dir = temp_save_dir
            else:
                # Clean up the temporary directory if we don't need to register to mlflow.
                if use_temp_dir:
                    shutil.rmtree(temp_save_dir)
        dist.barrier()

    def _save_and_register_peft_model(
        self,
        state: State,
        new_model_instance: Any,
        original_tokenizer: Optional[Any],
        save_dir: str,
    ):
        new_model_instance = self.transform_model_pre_registration(
            new_model_instance,
        )
        components = {'model': new_model_instance}
        if original_tokenizer is not None:
            components['tokenizer'] = original_tokenizer

        log.debug('Logging Hugging Face model to MLFlow')
        for i, mlflow_logger in enumerate(self.mlflow_loggers):
            log.debug(
                f'Registering model to UC at {mlflow_logger.model_registry_prefix}.{self.mlflow_registered_model_name}',
            )

            local_save_path = str(Path(save_dir) / f'mlflow_save_{i}',)

            # TODO: Remove after mlflow fixes the bug that makes this necessary
            import mlflow
            import mlflow.store
            mlflow.store._unity_catalog.registry.rest_store.get_feature_dependencies = lambda *args, **kwargs: ''

            model_saving_kwargs: dict[str, Any] = {
                'path': local_save_path,
            }
            model_saving_kwargs['flavor'] = 'peft'
            model_saving_kwargs['save_pretrained_dir'] = save_dir
            model_saving_kwargs['metadata'] = self.mlflow_logging_config[
                'metadata']

            context_manager = te.onnx_export(
                True,
            ) if is_te_imported and state.precision == Precision.AMP_FP8 else contextlib.nullcontext(
            )
            with context_manager:
                # Add the pip requirements directly to avoid mlflow
                # attempting to run inference on the model
                model_saving_kwargs['pip_requirements'] = [
                    'transformers',
                    'torch',
                ]
                mlflow_logger.save_model(**model_saving_kwargs)

            # Upload the license file generated by mlflow during the model saving.
            # Get and log the license file.
            license_filename = _maybe_get_license_filename(
                local_save_path,
                self.pretrained_model_name,
            )
            if license_filename is not None:
                mlflow_logger._mlflow_client.log_artifact(
                    mlflow_logger._run_id,
                    os.path.join(local_save_path, license_filename),
                )

            self.pre_register_edit(local_save_path)

            with _monitor_process_saver(mlflow_logger):
                process = SpawnProcess(
                    target=_register_model_with_run_id_multiprocess,
                    kwargs={
                        'mlflow_logger':
                            mlflow_logger,
                        'composer_logging_level':
                            logging.getLogger('composer').level,
                        'model_uri':
                            local_save_path,
                        'name':
                            self.mlflow_registered_model_name,
                        'await_creation_for':
                            3600,
                    },
                )
                process.start()
                self.register_processes.append(process)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Track training runs for loss spikes or persistently high training loss."""
from __future__ import annotations

import logging
from collections import deque

import numpy as np
import torch
from composer.core import Callback, State, TimeUnit
from composer.loggers import Logger, MosaicMLLogger
from composer.utils import dist

from llmfoundry.utils.exceptions import HighLossError, LossSpikeError
from llmfoundry.utils.warnings import experimental_class

log = logging.getLogger(__name__)

__all__ = ['KillLossSpike']

_MIN_WINDOW_SIZE = 100
_MAX_LOSS_CAP = 10
_WINDOW_FRACTION = 0.05


@experimental_class('KillLossSpike')
class KillLossSpike(Callback):
    """Detects and handles loss spikes or high losses during training.

    Monitors the training loss at the end of each batch and maintains a rolling window of recent losses.
    If recent training losses exceed a specified cap or if a significant spike in loss is detected, the callback can either
    log a warning (displayed as a message on the run event) or raise a LossSpikeError to stop the run without retry.

    Args:
        log_only (bool): If True, the callback will only log warnings without interrupting training. If False, a
                         LossSpikeError will be raised to stop training upon detecting a loss spike or persistently
                         high loss. Default is True.
        patience (int): The number of consecutive outlier losses tolerated before considering the training loss to be
                        persistently high. Default is 4 (so 5 consecutive outlier losses will trigger an error).
        outlier_multiplier (int): The multiplier used to determine if a loss is an outlier. A loss is considered an
                                  outlier if it is outlier_multiplier times greater than the mean of losses in
                                  the current window. Default is 2.
        window_size (int): The size of the rolling window used to track recent losses. This is set to 1/20 of the total training batches, with a minimum of 100 steps.
        loss_cap (int): The maximum allowable loss. If the training loss consistently exceeds this value,
                        it is considered a diverging or unstable run. This is set to the maximum loss from the first window of losses, with a maximum of 10.

    Raises:
        LossSpikeError: If log_only is False and a loss spike or persistently high loss is detected, this error is
                        raised to stop the run with an error message.
    """

    def __init__(
        self,
        log_only: bool = True,
        patience: int = 4,
        outlier_multiplier: float = 2,
        window_size: int = _MIN_WINDOW_SIZE,
        loss_cap: float = _MAX_LOSS_CAP,
    ):
        self._enabled = (dist.get_global_rank() == 0)
        self.log_only = log_only
        self.patience = patience
        self.outlier_multiplier = outlier_multiplier
        self.outlier_counter = 0
        self.window_size = window_size
        self.loss_window = deque()
        self.loss_cap = loss_cap
        self.window_size_set = (window_size != _MIN_WINDOW_SIZE)
        self.loss_cap_set = (loss_cap != _MAX_LOSS_CAP)

    def _detect_loss_spike(
        self,
        train_loss: float,
        running_loss_avg: float,
    ) -> bool:
        # Train loss is an outlier
        if train_loss >= running_loss_avg * self.outlier_multiplier:
            self.outlier_counter += 1
            log.info(
                f'Potential loss spike detected. Iteration: {self.outlier_counter}',
            )
            if self.outlier_counter > self.patience:
                log.info(
                    f'Loss spike detected for {self.outlier_counter} steps. Try lowering the learning rate.',
                )
                return True
        # Previous step loss was an outlier, current step loss is not. Reset outlier counter.
        elif self.outlier_counter > 0:
            log.info(f'Not a persistent loss spike. Resetting outlier counter.')
            self.outlier_counter = 0
        return False

    def _detect_high_losses(self, current_step: int) -> bool:
        if current_step < self.window_size * 2:
            return False

        # Half of the running losses are greater than our "high loss" threshold, after an initial buffer period
        high_loss_count = sum(
            1 for loss in self.loss_window if loss > self.loss_cap
        )
        is_high_loss = (high_loss_count >= self.window_size / 2)

        if is_high_loss:
            log.info(
                f'High losses detected: {high_loss_count}/{self.window_size} losses above {self.loss_cap}.',
            )

        return is_high_loss

    def _log_metadata(self, logger: Logger, key: str, value: dict) -> None:
        for destination in logger.destinations:
            if isinstance(destination, MosaicMLLogger):
                destination.log_metadata({
                    key: value,
                    'loss_window': list(self.loss_window),
                })

    def _handle_loss_spike(
        self,
        logger: Logger,
        running_loss_avg: float,
    ) -> None:
        if self.log_only:
            self._log_metadata(
                logger,
                'loss_spike',
                {
                    'outlier_multiplier': self.outlier_multiplier,
                    'running_loss_avg': running_loss_avg,
                    'outlier_counter': self.outlier_counter,
                },
            )
        else:
            raise LossSpikeError(
                outlier_multiplier=self.outlier_multiplier,
                running_loss_avg=running_loss_avg,
                outlier_counter=self.outlier_counter,
                loss_window=list(self.loss_window),
            )

    def _handle_high_losses(self, logger: Logger) -> None:
        if self.log_only:
            self._log_metadata(
                logger,
                'high_loss',
                {
                    'loss_cap': self.loss_cap,
                    'window_size': self.window_size,
                },
            )
        else:
            raise HighLossError(
                loss_cap=self.loss_cap,
                window_size=self.window_size,
                loss_window=list(self.loss_window),
            )

    def _set_window_size(self, state: State) -> None:
        total_training_steps = 0
        current_step = int(state.timestamp.batch)
        current_token = int(state.timestamp.token)

        if state.max_duration is not None:
            if state.max_duration.unit == TimeUnit.EPOCH and state.dataloader_len is not None:
                total_training_steps = state.dataloader_len * state.max_duration.value
            elif state.max_duration.unit == TimeUnit.BATCH:
                total_training_steps = state.max_duration.value
            elif state.max_duration.unit == TimeUnit.TOKEN:
                # This is an approximation of the total batches from the total tokens, assuming the ratio of tokens:batch is constant.
                total_training_steps = current_step * (
                    state.max_duration.value / current_token
                )
            self.window_size = max(
                self.window_size,
                round(float(total_training_steps * _WINDOW_FRACTION)),
            )
        self.loss_window = deque(self.loss_window, maxlen=self.window_size)
        log.info(f'Window size set to: {self.window_size}')
        self.window_size_set = True

    def batch_end(self, state: State, logger: Logger) -> None:

        if not isinstance(state.loss, torch.Tensor):
            raise NotImplementedError('Multiple losses not supported yet')
        train_loss = state.loss.item()

        # Only start early stopping once a full window of loss data
        if len(self.loss_window) == self.window_size:

            current_step = int(state.timestamp.batch)

            # If window size has not yet been set either by user or during run, set window size to a fraction of the total training duration. Minimum 100 batches.
            if not self.window_size_set:
                self._set_window_size(state)
                # Window size has been expanded -- keep adding losses until we reach the window size.
                if self.window_size > _MIN_WINDOW_SIZE:
                    self.loss_window.append(train_loss)
                    return

            # If user does not provide a loss cap, set loss cap to the maximum loss from the first loss window. Hard cap at loss=10.
            if not self.loss_cap_set and current_step == self.window_size:
                self.loss_cap = min(max(self.loss_window), self.loss_cap)
                self.loss_cap_set = True

            running_loss_avg = float(np.mean(self.loss_window))

            if self._detect_loss_spike(train_loss, running_loss_avg):
                self._handle_loss_spike(logger, running_loss_avg)
            elif self._detect_high_losses(current_step):
                self._handle_high_losses(logger)

        self.loss_window.append(train_loss)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import collections
from typing import Optional

__all__ = ['OutlierDetector']


class OutlierDetector:
    """OutlierDetector.

    This class implements an algorithm to detect outliers in sequential
    numeric data (e.g. for gradient/moment norms in optimizers). It relies on a
    delayed moving average which is the moving average of observations from time
    step T-2*`delay_interval` to T-`delay_interval`. The motivation is that
    outliers typically occur in clusters that can potentially stretch for many
    observations, hence it's best to use a delayed moving average to detect
    outliers.

    It defines an outlier as any data point that is `threshold` times larger than the delayed moving average.

    The class assumes data is inserted sequentially at evenly spaced intervals of unit length 1.
    """

    def __init__(self, threshold: float = 7.5, delay_interval: int = 500):

        self.intermediate_data_queue = collections.deque(maxlen=delay_interval)
        self.delayed_moving_average = collections.deque(maxlen=delay_interval)
        self.threshold = threshold

    def insert_observation(self, obs: float) -> bool:
        """Insert observation.

        Inserts obs into the data buffer and returns true if it is an "outlier", defined `threshold` times larger than
        the windowed moving average from [T-2*`delay_interval` : T-`delay_interval`].

        This algorithm first moves recent data into an intermediate circular buffer, and then moves data in to the delayed moving average buffer
        once it is old enough to be evicted from the intermediate data. This is to ensure that we take a delayed moving average that doesn't include recent data.

        Args:
            obs (float): Numeric observation for the current timestep.

        Returns:
            bool: Indicator of whether the most recent observation was an outlier.
        """
        assert self.intermediate_data_queue.maxlen is not None, 'expected maxlen defined'

        if len(
            self.intermediate_data_queue,
        ) >= self.intermediate_data_queue.maxlen:
            # move data from intermediate queue to slow moving average queue
            intermediate_obs = self.intermediate_data_queue.popleft()
            self.delayed_moving_average.append(intermediate_obs)

        self.intermediate_data_queue.append(obs)
        delayed_mva = self.get_delayed_mva()
        return delayed_mva is not None and obs > self.threshold * delayed_mva

    def get_delayed_mva(self) -> Optional[float]:
        if len(self.delayed_moving_average) > 0:
            return sum(
                self.delayed_moving_average,
            ) / len(self.delayed_moving_average)
        else:
            return None



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0
import json
import os
from typing import Any, Optional, Union

from composer.loggers import MosaicMLLogger
from composer.loggers.logger_destination import LoggerDestination
from composer.loggers.mosaicml_logger import (
    MOSAICML_ACCESS_TOKEN_ENV_VAR,
    MOSAICML_PLATFORM_ENV_VAR,
)

__all__ = [
    'maybe_create_mosaicml_logger',
    'find_mosaicml_logger',
    'log_eval_analytics',
    'log_train_analytics',
]

_MODEL_KEYS_TO_LOG = [
    'pretrained_model_name_or_path',
    'pretrained',
    'vocab_size',
    'd_model',
    'n_heads',
    'n_layers',
    'expansion_ratio',
    'max_seq_len',
]


def maybe_create_mosaicml_logger() -> Optional[MosaicMLLogger]:
    """Creates a MosaicMLLogger if the run was sent from the Mosaic platform."""
    if os.environ.get(MOSAICML_PLATFORM_ENV_VAR, 'false').lower(
    ) == 'true' and os.environ.get(MOSAICML_ACCESS_TOKEN_ENV_VAR):
        return MosaicMLLogger()


def find_mosaicml_logger(
    loggers: list[LoggerDestination],
) -> Optional[MosaicMLLogger]:
    """Returns the first MosaicMLLogger from a list, and None otherwise."""
    return next(
        (logger for logger in loggers if isinstance(logger, MosaicMLLogger)),
        None,
    )


def log_eval_analytics(
    mosaicml_logger: MosaicMLLogger,
    model_configs: list[dict[str, Any]],
    icl_tasks: Union[str, list[dict[str, Any]]],
    eval_gauntlet_config: Optional[Union[str, dict[str, Any]]],
):
    """Logs analytics for runs using the `eval.py` script."""
    metrics: dict[str, Any] = {
        'llmfoundry/script': 'eval',
    }

    metrics['llmfoundry/gauntlet_configured'] = eval_gauntlet_config is not None
    metrics['llmfoundry/icl_configured'] = isinstance(
        icl_tasks,
        str,
    ) or len(icl_tasks) > 0

    metrics['llmfoundry/model_configs'] = []
    for model_config in model_configs:
        nested_model_config = model_config.get('model', {})
        model_config_data = {}
        for key in _MODEL_KEYS_TO_LOG:
            if nested_model_config.get(key, None) is not None:
                model_config_data[key] = nested_model_config.get(key)

        if len(model_config_data) > 0:
            metrics['llmfoundry/model_configs'].append(
                json.dumps(model_config_data, sort_keys=True),
            )

    mosaicml_logger.log_metrics(metrics)
    mosaicml_logger._flush_metadata(force_flush=True)


def log_train_analytics(
    mosaicml_logger: MosaicMLLogger,
    model_config: dict[str, Any],
    train_loader_config: dict[str, Any],
    eval_loader_config: Optional[Union[dict[str, Any], list[dict[str, Any]]]],
    callback_configs: Optional[dict[str, Any]],
    tokenizer_name: str,
    load_path: Optional[str],
    icl_tasks_config: Optional[Union[list[dict[str, Any]], str]],
    eval_gauntlet: Optional[Union[dict[str, Any], str]],
):
    """Logs analytics for runs using the `train.py` script."""
    train_loader_dataset = train_loader_config.get('dataset', {})
    metrics: dict[str, Any] = {
        'llmfoundry/tokenizer_name': tokenizer_name,
        'llmfoundry/script': 'train',
        'llmfoundry/train_loader_name': train_loader_config.get('name'),
    }

    if callback_configs is not None:
        metrics['llmfoundry/callbacks'] = [
            name for name, _ in callback_configs.items()
        ]

    metrics['llmfoundry/gauntlet_configured'] = eval_gauntlet is not None
    metrics['llmfoundry/icl_configured'] = (
        icl_tasks_config is not None and
        ((isinstance(icl_tasks_config, str) or len(icl_tasks_config) > 0))
    )

    if train_loader_dataset.get('hf_name', None) is not None:
        metrics['llmfoundry/train_dataset_hf_name'] = train_loader_dataset.get(
            'hf_name',
            None,
        )
    if train_loader_config.get('name') == 'finetuning':
        metrics['llmfoundry/train_task_type'] = 'INSTRUCTION_FINETUNE'
    elif train_loader_config.get('name') == 'text':
        if load_path is not None or model_config.get('pretrained') == True:
            metrics['llmfoundry/train_task_type'] = 'CONTINUED_PRETRAIN'
        else:
            metrics['llmfoundry/train_task_type'] = 'PRETRAIN'

    if eval_loader_config is not None:
        metrics['llmfoundry/eval_loaders'] = []

        if isinstance(eval_loader_config, list):
            eval_loader_configs: list = eval_loader_config
        else:
            eval_loader_configs = [eval_loader_config]

        for loader_config in eval_loader_configs:
            eval_loader_info = {}
            eval_loader_dataset = loader_config.get('dataset', {})
            eval_loader_info['name'] = loader_config.get('name')
            if eval_loader_dataset.get('hf_name', None) is not None:
                eval_loader_info['dataset_hf_name'] = eval_loader_dataset.get(
                    'hf_name',
                )

            # Log as a key-sorted JSON string, so that we can easily parse it in Spark / SQL
            metrics['llmfoundry/eval_loaders'].append(
                json.dumps(eval_loader_info, sort_keys=True),
            )

    model_config_data = {}
    for key in _MODEL_KEYS_TO_LOG:
        if model_config.get(key, None) is not None:
            model_config_data[f'llmfoundry/{key}'] = model_config.get(key)

    if len(model_config_data) > 0:
        metrics.update(model_config_data)

    mosaicml_logger.log_metrics(metrics)
    mosaicml_logger._flush_metadata(force_flush=True)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry import callbacks, data, models, optim, tokenizers, utils


# This very simple test is just to use the above imports, which check and make sure we can import all the top-level
# modules from foundry. This is mainly useful for checking that we have correctly conditionally imported all optional
# dependencies.
def test_smoketest():
    assert callbacks
    assert data
    assert models
    assert optim
    assert tokenizers
    assert utils



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Streaming dataset conversion scripts for json files."""
import os
from enum import Enum
from glob import glob
from typing import Optional

import datasets as hf_datasets
from streaming import MDSWriter
from torch.utils.data import IterableDataset
from tqdm import tqdm
from transformers import AutoTokenizer, PreTrainedTokenizerBase

from llmfoundry.data import ConcatTokensDataset, NoConcatDataset


class ConcatMode(Enum):
    NO_CONCAT = 'NO_CONCAT'
    CONCAT_TOKENS = 'CONCAT_TOKENS'


def build_hf_dataset(
    path: str,
    split: str,
    mode: ConcatMode,
    max_length: Optional[int] = None,
    bos_text: str = '',
    eos_text: str = '',
    no_wrap: bool = False,
    tokenizer: PreTrainedTokenizerBase = None,
) -> IterableDataset:
    """Build an IterableDataset over the HF C4 or pile source data.

    Args:
        path (str): Dataset name
        split (str): Split name.
        mode (ConcatMode): NO_CONCAT, or CONCAT_TOKENS
        max_length (int): The length of concatenated tokens
        bos_text (str): text to insert at the beginning of each sequence
        eos_text (str): text to insert at the end of each sequence
        no_wrap (bool): if concatenating, whether to wrap text across `max_length` boundaries
        tokenizer (PreTrainedTokenizerBase): if mode is CONCAT_TOKENS, the tokenizer to use
        data_subset (str): Referred to as "name" in HuggingFace datasets.load_dataset.
            Typically "all" (The Pile) or "en" (allenai/c4).

    Returns:
        An IterableDataset.
    """
    if os.path.isdir(path):
        data_files = glob(f'{path}/*')
    else:
        data_files = path

    hf_dataset = hf_datasets.load_dataset(
        'json',
        data_files=data_files,
        split=split,
    )

    if mode == ConcatMode.NO_CONCAT:
        dataset = NoConcatDataset(hf_dataset)
    else:
        if not isinstance(tokenizer, PreTrainedTokenizerBase):
            raise ValueError(
                f'{tokenizer=} must be of type PreTrainedTokenizerBase',
            )
        if max_length is None:
            raise ValueError(f'max_length must be set.')
        if bos_text + eos_text == '':
            test_tokens = tokenizer('test')
            if test_tokens['input_ids'][
                0] != tokenizer.bos_token_id and test_tokens['input_ids'][
                    -1] != tokenizer.eos_token_id:
                tok_error_msg = 'This tokenizer does not insert an EOS nor BOS token. '
                tok_error_msg += 'Concatenating with this tokenizer will result in sequences being '
                tok_error_msg += 'attached without a separating token. Please use another tokenizer, '
                tok_error_msg += 'such as facebook/opt-125m, or specify EOS/BOS text with e.g. '
                tok_error_msg += '--bos_text=<|endoftext|>.'
                raise ValueError(tok_error_msg)
        dataset = ConcatTokensDataset(
            hf_dataset=hf_dataset,
            tokenizer=tokenizer,
            max_length=max_length,
            bos_text=bos_text,
            eos_text=eos_text,
            no_wrap=no_wrap,
        )
    return dataset


def convert_dataset_json(
    path: str,
    out_root: str,
    compression: Optional[str],
    concat_tokens: Optional[int],
    split: str,
    tokenizer: Optional[str] = None,
    bos_text: str = '',
    eos_text: str = '',
    no_wrap: bool = False,
    num_workers: Optional[int] = None,
) -> None:
    """Create C4/pile streaming dataset.

    Args:
        path (str): Path to the input data file
        out_root (str): Output root directory
        compression (Optional[str]): Compression type, if any
        concat_tokens (Optional[int]): Convert text to tokens and concatenate up to this many tokens
        split (str): Dataset split to process
        tokenizer (Optional[str]): Tokenizer name
        bos_text (str): Text to insert at the beginning of each sequence
        eos_text (str): Text to insert at the end of each sequence
        no_wrap (bool): Do not wrap text across max_length boundaries
        num_workers (Optional[int]): Number of workers for data loading
    """
    if concat_tokens is not None:
        mode = ConcatMode.CONCAT_TOKENS
        built_tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        # we will enforce length, so suppress warnings about sequences too long for the model
        built_tokenizer.model_max_length = int(1e30)
        columns = {'tokens': 'ndarray:int32'}
    else:
        mode = ConcatMode.NO_CONCAT
        built_tokenizer = None
        columns = {'text': 'str'}

    # Get samples
    dataset = build_hf_dataset(
        path=path,
        split=split,
        mode=mode,
        max_length=concat_tokens,
        bos_text=bos_text,
        eos_text=eos_text,
        no_wrap=no_wrap,
        tokenizer=built_tokenizer,
    )

    print('here')

    # Write samples
    print(f'Converting to MDS format...')
    print(
        f'Note that the progress bar is based on the dataset length before tokenization.',
    )
    print(f'It will finish at a value below 100% if tokenizing')
    with MDSWriter(
        columns=columns,
        out=os.path.join(out_root),
        compression=compression,
    ) as out:
        for sample in tqdm(dataset):
            out.write(sample)


def convert_dataset_json_from_args(
    path: str,
    out_root: str,
    compression: Optional[str],
    concat_tokens: Optional[int],
    split: str,
    tokenizer: Optional[str] = None,
    bos_text: Optional[str] = None,
    eos_text: Optional[str] = None,
    no_wrap: bool = False,
    num_workers: Optional[int] = None,
) -> None:
    """A wrapper for `convert_dataset_json` that parses arguments.

    Args:
        path (str): Path to the input data file
        out_root (str): Output root directory
        compression (Optional[str]): Compression type, if any
        concat_tokens (Optional[int]): Convert text to tokens and concatenate up to this many tokens
        split (str): Dataset split to process
        tokenizer (Optional[str]): Tokenizer name
        bos_text (Optional[str]): Text to insert at the beginning of each sequence
        eos_text (Optional[str]): Text to insert at the end of each sequence
        no_wrap (bool): Do not wrap text across max_length boundaries
        num_workers (Optional[int]): Number of workers for data loading

    Raises:
        ValueError: If the out_root directory exists and contains files that overlap with the requested splits
        ValueError: If concat_tokens is set and a tokenizer is not provided
    """
    os.environ['WORLD_SIZE'] = '1'
    if os.path.isdir(out_root) and len(
        set(os.listdir(out_root)).intersection(set(split)),
    ) > 0:
        raise ValueError(
            f'--out_root={out_root} contains {os.listdir(out_root)} which cannot overlap with the requested splits {split}.',
        )

    # Make sure we have needed concat options
    if (
        concat_tokens is not None and isinstance(concat_tokens, int) and
        tokenizer is None
    ):
        ValueError(
            'When setting --concat_tokens, you must specify a --tokenizer',
        )

    # now that we have validated them, change BOS/EOS to strings
    if bos_text is None:
        bos_text = ''
    if eos_text is None:
        eos_text = ''

    convert_dataset_json(
        path=path,
        out_root=out_root,
        compression=compression,
        concat_tokens=concat_tokens,
        split=split,
        tokenizer=tokenizer,
        bos_text=bos_text,
        eos_text=eos_text,
        no_wrap=no_wrap,
        num_workers=num_workers,
    )



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging
import warnings
from typing import Any, Optional, Union

import torch
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

from llmfoundry.utils.consts import CROSS_ENTROPY_IGNORE_INDEX

log = logging.getLogger(__name__)

__all__ = [
    'Seq2SeqFinetuningCollator',
]

TokenizedExample = dict[str, list[dict[str, list[int]]]]


def ensure_list(x: Union[list, torch.Tensor]) -> list:
    if isinstance(x, torch.Tensor):
        x = list(x.flatten())
    assert isinstance(x, list)
    return x


def validate_target_settings(
    target_prompts: str,
    target_responses: str,
    decoder_only_format: bool,
):
    """Raises an error if target settings are invalid."""
    if (not decoder_only_format
       ) and (target_prompts != 'none' or target_responses != 'last'):
        raise ValueError(
            f'When using encoder_decoder format, you must use target_prompts="none" and target_responses="last".',
        )

    if target_responses not in {'all', 'last'}:
        raise ValueError(
            f'target_responses must be either "last" or "all" but {target_responses=}',
        )

    if target_prompts.startswith('length>='):
        cutoff = target_prompts[8:]
        if not cutoff.isdigit():
            raise ValueError(
                f'target_prompts starts with "length>=" but the rest of the string is not digits ({target_prompts=}). ' +\
                'To use this configuration option, set target_prompts "length>=XX" where "XX" is a positive integer indicating ' +\
                'the length cutoff. Prompts of at least XX tokens in length will be treated as targets.',
            )
        cutoff = int(cutoff)
        if cutoff <= 0:
            raise ValueError(
                f'You are trying to set the target_prompts length cutoff to a negative number {cutoff=}. This is not allowed.',
            )
    elif target_prompts not in {'all', 'none'}:
        raise ValueError(
            f'target_prompts must either be "all", "none" or "length>=XX" where "XX" is a positive integer, but {target_prompts=}',
        )


###### Functions to implement target_prompts and target_responses choices #####
def _sequence_to_labels_all(
    sequence: list[int],
    is_last_turn: bool,
    cutoff: Optional[int] = None,
) -> list[int]:
    del is_last_turn, cutoff  # unused
    return sequence


def _sequence_to_labels_none(
    sequence: list[int],
    is_last_turn: bool,
    cutoff: Optional[int] = None,
) -> list[int]:
    del is_last_turn, cutoff  # unused
    return [CROSS_ENTROPY_IGNORE_INDEX] * len(sequence)


def _sequence_to_labels_last(
    sequence: list[int],
    is_last_turn: bool,
    cutoff: Optional[int] = None,
) -> list[int]:
    del cutoff  # unused
    if is_last_turn:
        return sequence
    else:
        return [CROSS_ENTROPY_IGNORE_INDEX] * len(sequence)


def _sequence_to_labels_cutoff(
    sequence: list[int],
    is_last_turn: bool,
    cutoff: Optional[int] = None,
) -> list[int]:
    del is_last_turn  # unused
    if cutoff is None:
        raise ValueError('input ``cutoff`` must be provided')
    if len(sequence) >= cutoff:
        return sequence
    else:
        return [CROSS_ENTROPY_IGNORE_INDEX] * len(sequence)


_TARGET_POLICY_LOOKUP = {
    'all': _sequence_to_labels_all,
    'none': _sequence_to_labels_none,
    'last': _sequence_to_labels_last,
    'length': _sequence_to_labels_cutoff,
}


def stitch_turns_decoder_only(
    example_turns: list[dict[str, list[int]]],
    target_prompts: str,
    target_responses: str,
    eos_token_id: Optional[int] = None,
    validate: bool = False,
) -> tuple[list[int], list[int]]:
    target_prompts = target_prompts.lower()
    target_responses = target_responses.lower()

    if validate:
        validate_target_settings(
            target_prompts,
            target_responses,
            decoder_only_format=True,
        )

    if target_prompts.startswith('length'):
        prompt_cutoff = int(target_prompts.split('>=')[-1])
        prompt_to_target = _TARGET_POLICY_LOOKUP['length']
    else:
        prompt_cutoff = None
        prompt_to_target = _TARGET_POLICY_LOOKUP[target_prompts]
    response_to_target = _TARGET_POLICY_LOOKUP[target_responses]

    input_ids = []
    labels = []
    for idx, turn in enumerate(example_turns):
        is_last_turn = idx + 1 == len(example_turns)
        # We assume that no padding has been applied. If there is a pad token,
        # it may be because the pad token is the same as another special token.
        context = ensure_list(turn['input_ids'])
        target = ensure_list(turn['labels'])
        # If an EOS token id is given, ensure that the target sequence ends with it.
        if is_last_turn and eos_token_id is not None:
            if target[-1] != eos_token_id:
                target = target + [eos_token_id]
        # Extend the input_ids
        input_ids += context
        input_ids += target
        # Extend the labels, with values depending on the loss-generating policies
        labels += prompt_to_target(context, is_last_turn, prompt_cutoff)
        labels += response_to_target(target, is_last_turn)

    if len(input_ids) != len(labels):
        raise ValueError(
            f'input_ids and labels should be the same length, {len(input_ids)=}, {len(labels)=}',
        )
    return input_ids, labels


def stitch_turns_encoder_decoder(
    example_turns: list[dict[str, list[int]]],
    eos_token_id: Optional[int] = None,
) -> tuple[list[int], list[int]]:
    context = []
    target = None
    for idx, turn in enumerate(example_turns):
        is_last_turn = idx + 1 == len(example_turns)
        # We assume that no padding has been applied. If there is a pad token,
        # it may be because the pad token is the same as another special token.
        turn_context = ensure_list(turn['input_ids'])
        turn_target = ensure_list(turn['labels'])
        # Context always goes into input_ids.
        context += turn_context
        # Non-terminal turns go into the input_ids. Terminal target is the target.
        if is_last_turn:
            # If an EOS token id is given, ensure that the target sequence ends with it.
            if eos_token_id is not None and turn_target[-1] != eos_token_id:
                turn_target = turn_target + [eos_token_id]
            target = turn_target
        else:
            context += turn_target
    if target is None:
        raise ValueError('target is still None but should be list[int]')
    return context, target


class Seq2SeqFinetuningCollator:
    """A general-purpose collator for sequence-to-sequence training/evaluation.

    Args:
        tokenizer: A HuggingFace tokenizer. Must have a pad_token set.
        max_seq_len (int): The maximum sequence length of the combined
            context/target sequence (decoder-only format) or of each the
            context sequence and target sequence (encoder-decoder format).
        decoder_only_format (bool): Whether to format the batches for a
            decoder-only model (if True) or an encoder-decoder model (if False).
        target_responses (str): For multi-turn examples, this controls which
            responses are treated as training targets (i.e. generate loss).
            Options are:
                "last": (Default) Only the final response is used as the training
                    target; non-terminal responses are only part of the context.
                "all": All of the responses are used as training targets.
        target_prompts (str): This controls which prompts are treated as
            training targets (i.e. generate loss).
            Options are:
                "none": (Default) Prompts are never used as training targets.
                "all": Prompts are always used as training targets.
                "length>=XX": Prompt sequences are used as training targets when
                    they have length of at least XX tokens. For instance,
                    setting "length>=512" instructs the collator to use a prompt
                    sequence as a training target when it is at least 512 tokens long.
        allow_pad_trimming (bool, optional): Whether to allow the collator
            to trim padding, which may result in smaller but inconsistent batch
            sizes. Default: ``False`` ensures that all sequences are max_seq_len.
        batch_metadata (dict, optional): A dictionary of metadata which will be added
            to the batch.
        pad_to_longest (bool, optional): Whether to pad to the longest sequence,
            which may result in smaller but inconsistent batch sizes. This is
            primarily used to profile packing.
            Default: ``False`` ensures that all sequences are max_seq_len.
    """

    def __init__(
        self,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        max_seq_len: int,
        decoder_only_format: bool,
        target_responses: str = 'last',
        target_prompts: str = 'none',
        allow_pad_trimming: bool = False,
        batch_metadata: Optional[dict[str, Any]] = None,
        pad_to_longest: bool = False,
    ):
        self.tokenizer = tokenizer
        self.max_seq_len = max_seq_len
        self.decoder_only_format = decoder_only_format
        self.target_responses = target_responses.lower()
        self.target_prompts = target_prompts.lower()
        self.batch_metadata = batch_metadata or {}

        # Trimming will always be skipped on at least the first __call__
        self._allow_pad_trimming = allow_pad_trimming
        self._seen_first_batch = False

        self._pad_to_longest = pad_to_longest

        illegal_keys = [
            'input_ids',
            'labels',
            'attention_mask',
            'decoder_input_ids',
            'decoder_attention_mask',
        ]
        found_keys = [
            illegal_key for illegal_key in illegal_keys
            if illegal_key in self.batch_metadata
        ]
        if found_keys:
            raise ValueError(
                f'The following keys are in batch_metadata but are not allowed: {", ".join(found_keys)}.\n' +\
                f'You cannot use keys that are used directly by the models. The prohibited keys are:\n' +\
                f'{", ".join(illegal_keys)}',
            )

        if (max_seq_len % 8) != 0:
            log.warning(
                'For performance, a max_seq_len as a multiple of 8 is recommended.',
            )

        if self.tokenizer.pad_token_id is None:
            raise ValueError(
                f'{self.__class__.__name__} requires that the tokenizer has the pad token set, but it is None',
            )

        validate_target_settings(
            self.target_prompts,
            self.target_responses,
            self.decoder_only_format,
        )
        if self.target_prompts.startswith('length'):
            self.prompt_cutoff = int(self.target_prompts.split('>=')[-1])
            self.prompt_to_target = _TARGET_POLICY_LOOKUP['length']
        else:
            self.prompt_cutoff = None
            self.prompt_to_target = _TARGET_POLICY_LOOKUP[self.target_prompts]
        self.response_to_target = _TARGET_POLICY_LOOKUP[self.target_responses]

        self._warned_truncated = False
        self._warned_context = False
        self._warned_target = False

    def __call__(self,
                 examples: list[TokenizedExample]) -> dict[str, torch.Tensor]:
        for check_key in ['input_ids', 'labels']:
            if check_key not in examples[0]['turns'][0]:
                raise KeyError(
                    f'Examples returned by dataset do not include required key: {check_key}',
                )

        if self.decoder_only_format:
            batch = self._process_and_batch_decoder_only(examples)
        else:
            batch = self._process_and_batch_encoder_decoder(examples)

        # Add any batch_metadata
        batch_size = batch['input_ids'].shape[0]
        batch.update({
            k: torch.tensor([v] * batch_size)
            for k, v in self.batch_metadata.items()
        })

        return batch

    def _process_and_batch_decoder_only(
        self,
        examples: list[TokenizedExample],
    ) -> dict[str, torch.Tensor]:
        # Steps explained in comments
        processed_examples = []
        input_ids_and_labels = [
            stitch_turns_decoder_only(
                example_turns=example['turns'],
                target_prompts=self.target_prompts,
                target_responses=self.target_responses,
                eos_token_id=self.tokenizer.eos_token_id,
            ) for example in examples
        ]

        if self._pad_to_longest:
            max_seq_len = max([
                len(input_ids) for input_ids, _ in input_ids_and_labels
            ])
            max_seq_len = min(max_seq_len, self.max_seq_len)
        else:
            max_seq_len = self.max_seq_len

        for input_ids, labels in input_ids_and_labels:
            orig_size = len(input_ids)
            # We may need to truncate the input_ids / labels in order to maintain max_seq_len
            if orig_size > max_seq_len:
                input_ids = input_ids[:max_seq_len]
                labels = labels[:max_seq_len]

                # Check to make sure there are still loss-generating tokens. Error if not.
                if len([l for l in labels if l != CROSS_ENTROPY_IGNORE_INDEX
                       ],) == 0:
                    raise ValueError(
                        f'Truncating to max_seq_len={max_seq_len} has removed all loss-generating tokens. ' +\
                        f'Pre-truncation sequence length was {orig_size}. ' +\
                        'This sample should have been filtered out before reaching the collator. If using ' +\
                        'pre-tokenized streaming data, this may have resulted from using different ' +\
                        '``target_prompts``, ``target_responses``, or ``max_seq_len`` ' +\
                        'settings when preparing the streaming dataset than what are currently being used.',
                    )

                # Still issue a warning when truncating
                if not self._warned_truncated:
                    warnings.warn(
                        f'Truncating sequence of length={orig_size} to fit max_seq_len={max_seq_len}. ' +\
                        f'If truncation is a problem, consider increasing max_seq_len.',
                    )
                    self._warned_truncated = True

            attention_mask = [1] * len(input_ids)

            # Annoyingly, we need to pad everything but input_ids
            # and attention_mask ourselves
            n_total = len(input_ids)
            i_pad = [CROSS_ENTROPY_IGNORE_INDEX] * (max_seq_len - n_total)
            if self.tokenizer.padding_side == 'left':
                labels = i_pad + labels
            else:
                labels = labels + i_pad

            # Update the example
            processed_example = {
                'input_ids': input_ids,
                'labels': labels,
                'attention_mask': attention_mask,
            }

            processed_examples.append(processed_example)

        batch = self.tokenizer.pad(
            processed_examples,
            padding='max_length',
            max_length=max_seq_len,
            return_tensors='pt',
        )

        batch['sequence_id'] = batch['attention_mask'] - 1

        # This logic prevents trimming on at least the first batch
        if not (self._allow_pad_trimming and self._seen_first_batch):
            self._seen_first_batch = True
            return batch
        self._seen_first_batch = True

        # The batch is ready, but we can trim padding for efficiency
        multiple_of = 8

        n_non_padding = batch['attention_mask'].sum(dim=1).max()
        keep_tokens = int(multiple_of * torch.ceil(n_non_padding / multiple_of))
        for k, v in batch.items():
            if len(v.shape) < 2:
                continue
            if self.tokenizer.padding_side == 'left':
                batch[k] = v[:, -keep_tokens:].contiguous()
            else:
                batch[k] = v[:, :keep_tokens].contiguous()

        return batch

    def _process_and_batch_encoder_decoder(
        self,
        examples: list[TokenizedExample],
    ) -> dict[str, torch.Tensor]:
        # The encoder-decoder case is has some gotchas.
        # Steps are explained in comments.
        processed_examples = []
        contexts_and_targets = [
            stitch_turns_encoder_decoder(
                example_turns=example['turns'],
                eos_token_id=self.tokenizer.eos_token_id,
            ) for example in examples
        ]

        if self._pad_to_longest:
            max_seq_len = 0
            for context, target in contexts_and_targets:
                max_seq_len = max(max_seq_len, len(context), len(target))
        else:
            max_seq_len = self.max_seq_len

        for context, target in contexts_and_targets:
            # We need to pad labels ourselves. Because HF.
            if len(target) < max_seq_len:
                i_pad = [
                    CROSS_ENTROPY_IGNORE_INDEX,
                ] * (max_seq_len - len(target))
                target = target + i_pad
            else:
                if not self._warned_target:
                    warnings.warn(
                        f'Truncating TARGET sequence of length={len(target)} ' +\
                        f'to max_seq_len={max_seq_len}. If truncation is ' +\
                        f'a problem, consider increasing max_seq_len.')
                    self._warned_target = True
                target = target[:max_seq_len -
                                1] + [self.tokenizer.eos_token_id]

            # We might need to truncate the context. Preserve the beginning.
            if len(context) > max_seq_len:
                if not self._warned_context:
                    warnings.warn(
                        f'Truncating CONTEXT sequence of length={len(context)} ' +\
                        f'to max_seq_len={max_seq_len}. If truncation is ' +\
                        f'a problem, consider increasing max_seq_len.')
                    self._warned_context = True
                context = context[:max_seq_len -
                                  1] + [self.tokenizer.eos_token_id]

            # Back into the example
            processed_example = {
                'input_ids': context,
                'labels': target,
                'attention_mask': [1] * len(context),
            }

            processed_examples.append(processed_example)

        # Batch examples into a single dict (this also pads)
        batch = self.tokenizer.pad(
            processed_examples,
            padding='max_length',
            max_length=max_seq_len,
            return_tensors='pt',
        )
        # We're still missing decoder_input_ids and decoder_attention_mask
        batch['decoder_input_ids'] = torch.cat([
            torch.full((len(processed_examples), 1),
                       self.tokenizer.pad_token_id),
            batch['labels'][:, :-1],
        ],
                                               dim=1)
        batch['decoder_input_ids'].masked_fill_(
            batch['decoder_input_ids'] == CROSS_ENTROPY_IGNORE_INDEX,
            self.tokenizer.pad_token_id,
        )
        batch['decoder_attention_mask'] = torch.not_equal(
            batch['labels'],
            CROSS_ENTROPY_IGNORE_INDEX,
        )

        # This logic prevents trimming on at least the first batch
        if not (self._allow_pad_trimming and self._seen_first_batch):
            self._seen_first_batch = True
            return batch
        self._seen_first_batch = True

        # The batch is now valid, but we can trim padding for efficiency
        multiple_of = 8
        # (first for the encoder)
        n_non_padding = batch['attention_mask'].sum(dim=1).max()
        keep_tokens = int(multiple_of * torch.ceil(n_non_padding / multiple_of))
        for k in ['input_ids', 'attention_mask']:
            batch[k] = batch[k][:, :keep_tokens].contiguous()
        # (then for the decoder)
        n_non_padding = batch['decoder_attention_mask'].sum(dim=1).max()
        keep_tokens = int(multiple_of * torch.ceil(n_non_padding / multiple_of))
        for k in ['decoder_input_ids', 'decoder_attention_mask', 'labels']:
            batch[k] = batch[k][:, :keep_tokens].contiguous()

        return batch



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.data.data import (
    SUPPORTED_MDS_ENCODING_TYPES,
    ConcatTokensDataset,
    NoConcatDataset,
    stream_remote_local_validate,
)
from llmfoundry.data.dataloader import build_dataloader
from llmfoundry.data.finetuning import (
    Seq2SeqFinetuningCollator,
    StreamingFinetuningDataset,
    build_finetuning_dataloader,
)
from llmfoundry.data.packing import (
    BinPackCollator,
    auto_packing_ratio,
    profile_packing,
)
from llmfoundry.data.text_data import (
    ConcatenatedSequenceCollatorWrapper,
    StreamingTextDataset,
    build_text_dataloader,
)
from llmfoundry.data.utils import (
    get_data_spec,
    get_finetuning_collator,
    get_text_collator,
    validate_ds_replication,
)
from llmfoundry.registry import (
    collators,
    data_specs,
    dataloaders,
    dataset_replication_validators,
)

dataloaders.register('text', func=build_text_dataloader)
dataloaders.register('finetuning', func=build_finetuning_dataloader)

dataset_replication_validators.register(
    'dataset_replication_validator',
    func=validate_ds_replication,
)
collators.register('finetuning_collator', func=get_finetuning_collator)
collators.register('text_collator', func=get_text_collator)
data_specs.register('data_spec', func=get_data_spec)

__all__ = [
    'Seq2SeqFinetuningCollator',
    'build_finetuning_dataloader',
    'StreamingFinetuningDataset',
    'StreamingTextDataset',
    'build_text_dataloader',
    'NoConcatDataset',
    'ConcatTokensDataset',
    'build_dataloader',
    'BinPackCollator',
    'auto_packing_ratio',
    'profile_packing',
    'ConcatenatedSequenceCollatorWrapper',
    'stream_remote_local_validate',
    'SUPPORTED_MDS_ENCODING_TYPES',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Experimental learning rate schedulers used for training LLMs."""

import textwrap
import warnings
from typing import Union

from composer.core import State, Time, TimeUnit
from composer.optim import ComposerScheduler, LinearScheduler
from composer.optim.scheduler import _convert_time

from llmfoundry.utils.warnings import experimental_class

__all__ = ['InverseSquareRootWithWarmupScheduler']


def _raise_if_units_dont_match(
    time: Union[str, Time],
    t_max: Union[str, Time],
    name: str,
) -> None:
    new_time = Time.from_input(time)
    new_t_max = Time.from_input(t_max)

    if new_time.unit != new_t_max.unit:
        raise ValueError(
            f'{name} (unit {new_time.unit=}) must match max_duration unit ({new_t_max.unit=}).',
        )


def _raise_if_units_dur(time: Union[str, Time], name: str) -> None:
    new_time = Time.from_input(time)

    if new_time.unit == TimeUnit('dur'):
        raise ValueError(f'{name} cannot be in units of "dur".')


@experimental_class('InverseSquareRootWithWarmupScheduler')
class InverseSquareRootWithWarmupScheduler(ComposerScheduler):
    r"""Inverse square root LR decay with warmup and optional linear cooldown.

    Specifically, the learning rate multiplier :math:`\alpha(t)` can be expressed as:

    .. math::
        \alpha(t) = \begin{cases}
            t / t_{warmup}, & \text{if } t < t_{warmup} \\
            \alpha_{f,decay} + \frac{1 - \alpha_{f,decay}}{\sqrt{\tau_d}}, & \text{if } t_{warmup} <= t < t_{max} - t_{cooldown} \\
            \alpha_i + (alpha_{f,cooldown} - \alpha_i) \times \tau_c, & \text{otherwise}
        \end{cases}

    Given :math:`\tau_d`, the time elapsed during the inverse square root decay (normalized by :math:`t_scale`), as:

    .. math::
        \tau_d = (t - t_{warmup} + t_{scale}) / {t_scale}

    :math:`\alpha_i` as the value of the learning rate multiplier when :math:`\tau_d` is evaluated at :math:`t = t_{max} - t_{cooldown}`,
    and :math:`\tau_c`, the fraction of linear cooldown time elapsed (clipped to the interval :math:`[0, 1]`), as:

    .. math::
        \tau_c = (t - t_{max} + t_{cooldown}) / t_{cooldown}

    Where :math:`t_{warmup}` represents the warmup time, :math:`t_{scale}` represents the time scale,
    :math:`t_{cooldown}` represents the cooldown time, :math:`t_{max}` represents the duration of this scheduler,
    :math:`\alpha_{f,decay}` represents the learning rate multiplier that the inverse square root decays to at infinite time,
    and :math:`\alpha_{f,cooldown}` represents the learning rate multiplier that the linear cooldown decays to.

    Note, :math:`\alpha_{f,decay} >= \alpha_{f,cooldown}` to ensure that the learning rate is monotonically decreasing after warmup.

    Also note, ``t_warmup``, ``t_scale``, and ``t_cooldown`` cannot be specified in units of duration; since this schedule is designed for continual learning,
    ``max_duration`` is expected to change. Instead, these parameters need to be specified in the same units as ``max_duration`` passed to the trainer.

    Args:
        t_warmup (str | Time): The warmup time.
        t_scale (str | Time): The time scale.
        t_cooldown (str | Time): The cooldown time.
        t_max (str | Time): The duration of this scheduler. Default = ``"1dur"``.
        alpha_f_decay (float): The learning rate multiplier to decay inverse square root decay to. Default = ``0.0``.
        alpha_f_cooldown (float): The learning rate multiplier to decay linear cooldown to. Default = ``0.0``.
    """

    def __init__(
        self,
        t_warmup: Union[str, Time],
        t_scale: Union[str, Time],
        t_cooldown: Union[str, Time],
        t_max: Union[str, Time] = '1dur',
        alpha_f_decay: float = 0.0,
        alpha_f_cooldown: float = 0.0,
    ) -> None:
        if alpha_f_decay < alpha_f_cooldown:
            raise ValueError((
                'Required: alpha_f_decay >= alpha_f_cooldown. '
                f'Current: alpha_f_decay={alpha_f_decay}, '
                f'alpha_f_cooldown={alpha_f_cooldown}.'
            ))
        _raise_if_units_dur(t_warmup, 't_warmup')
        _raise_if_units_dur(t_scale, 't_scale')
        _raise_if_units_dur(t_cooldown, 't_cooldown')
        self.t_warmup = t_warmup
        self.t_scale = t_scale
        self.t_cooldown = t_cooldown
        self.t_max = t_max
        self.alpha_f_decay = alpha_f_decay
        self.alpha_f_cooldown = alpha_f_cooldown
        self.warmup_scheduler = LinearScheduler(
            alpha_i=0.0,
            alpha_f=1.0,
            t_max=t_warmup,
        )

    def __call__(self, state: State, ssr: float = 1.0) -> float:
        assert state.max_duration is not None, 'max_duration should be set whenever schedulers are invoked'
        _raise_if_units_dont_match(
            self.t_warmup,
            state.max_duration,
            't_warmup',
        )
        _raise_if_units_dont_match(self.t_scale, state.max_duration, 't_scale')
        _raise_if_units_dont_match(
            self.t_cooldown,
            state.max_duration,
            't_cooldown',
        )

        t_warmup = _convert_time(self.t_warmup, state)
        if t_warmup.value == 0:
            warnings.warn(
                textwrap.dedent(
                    """\
                The warmup duration is 0. If warmup was specified as a fraction of the total
                training duration, the warmup duration is calculated in the
                same unit as the trainer's max_duration parameter.""",
                ),
            )

        if state.timestamp < t_warmup:
            return self.warmup_scheduler(state)

        t_scale = _convert_time(self.t_scale, state, ssr=ssr)
        t_cooldown = _convert_time(self.t_cooldown, state, ssr=ssr)
        t_max = _convert_time(self.t_max, state, ssr=ssr)
        current_time = state.timestamp.get(t_scale.unit)

        t_shift = t_scale - t_warmup
        # t_cooldown_start is max of t_warmup, t_max - t_cooldown
        t_cooldown_start = t_max - t_cooldown
        if t_cooldown_start < t_warmup:
            t_cooldown_start = t_warmup

        if state.timestamp < t_cooldown_start:
            # Rescale LR by a coefficient equal to the inverse square root of the time
            # elapsed after warmup, rescaled by the time scale, such that, at
            # infinite time, the LR decays to alpha_f_decay.
            coeff = 1 / ((current_time + t_shift) / t_scale).value**0.5
            current_factor = (
                self.alpha_f_decay + coeff * (1.0 - self.alpha_f_decay)
            )
            return current_factor

        else:
            coeff = 1 / ((t_cooldown_start + t_shift) / t_scale).value**0.5
            alpha_i = self.alpha_f_decay + coeff * (1.0 - self.alpha_f_decay)

            if t_cooldown.value == 0:
                return alpha_i

            # Linearly decay the LR from its value at the step at which cooldown
            # started to alpha_f_cooldown over t_cooldown time.
            frac_of_cooldown = ((current_time - t_cooldown_start) /
                                t_cooldown).value
            frac_of_cooldown = min(1.0, frac_of_cooldown)
            current_factor = (
                alpha_i + frac_of_cooldown * (self.alpha_f_cooldown - alpha_i)
            )
            return current_factor



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.registry import config_transforms
from llmfoundry.utils.builders import (
    add_metrics_to_eval_loaders,
    build_algorithm,
    build_callback,
    build_composer_model,
    build_eval_loaders,
    build_evaluators,
    build_icl_data_and_gauntlet,
    build_icl_evaluators,
    build_logger,
    build_metric,
    build_optimizer,
    build_scheduler,
    build_tokenizer,
)
from llmfoundry.utils.checkpoint_conversion_helpers import (
    convert_and_save_ft_weights,
    get_hf_tokenizer_from_composer_state_dict,
    load_tokenizer,
)
from llmfoundry.utils.config_utils import (
    calculate_batch_size_info,
    log_config,
    pop_config,
    process_init_device,
    update_batch_size_info,
)
from llmfoundry.utils.consts import CROSS_ENTROPY_IGNORE_INDEX
from llmfoundry.utils.data_prep_utils import (
    DownloadingIterable,
    merge_shard_groups,
)
from llmfoundry.utils.huggingface_hub_utils import \
    edit_files_for_hf_compatibility
from llmfoundry.utils.logging_utils import SpecificWarningFilter
from llmfoundry.utils.model_download_utils import (
    download_from_hf_hub,
    download_from_http_fileserver,
    download_from_oras,
)
from llmfoundry.utils.mosaicml_logger_utils import (
    find_mosaicml_logger,
    log_eval_analytics,
    log_train_analytics,
    maybe_create_mosaicml_logger,
)
from llmfoundry.utils.prompt_files import load_prompts, load_prompts_from_file
from llmfoundry.utils.registry_utils import (
    TypedRegistry,
    construct_from_registry,
    create_registry,
    import_file,
    save_registry,
)
from llmfoundry.utils.warnings import (
    ExperimentalWarning,
    VersionedDeprecationWarning,
    experimental_class,
    experimental_function,
)

config_transforms.register(
    'update_batch_size_info',
    func=update_batch_size_info,
)

__all__ = [
    'add_metrics_to_eval_loaders',
    'build_algorithm',
    'build_callback',
    'build_eval_loaders',
    'build_evaluators',
    'build_icl_data_and_gauntlet',
    'build_icl_evaluators',
    'build_logger',
    'build_optimizer',
    'build_scheduler',
    'build_tokenizer',
    'build_composer_model',
    'build_metric',
    'get_hf_tokenizer_from_composer_state_dict',
    'load_tokenizer',
    'convert_and_save_ft_weights',
    'pop_config',
    'calculate_batch_size_info',
    'update_batch_size_info',
    'process_init_device',
    'log_config',
    'DownloadingIterable',
    'merge_shard_groups',
    'edit_files_for_hf_compatibility',
    'SpecificWarningFilter',
    'download_from_http_fileserver',
    'download_from_hf_hub',
    'download_from_oras',
    'maybe_create_mosaicml_logger',
    'find_mosaicml_logger',
    'log_eval_analytics',
    'log_train_analytics',
    'load_prompts',
    'load_prompts_from_file',
    'TypedRegistry',
    'create_registry',
    'construct_from_registry',
    'import_file',
    'save_registry',
    'VersionedDeprecationWarning',
    'ExperimentalWarning',
    'experimental_function',
    'experimental_class',
    'CROSS_ENTROPY_IGNORE_INDEX',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import math
import warnings
from collections.abc import Sequence
from copy import deepcopy
from functools import partial
from typing import Any, Callable, Optional, Union

import torch
from torch import nn
from torch.distributed._tensor import DTensor

from llmfoundry.layers_registry import (
    fcs,
    module_init_fns,
    norms,
    param_init_fns,
)
from llmfoundry.models.layers.dmoe import GLU, MLP

try:
    import transformer_engine.pytorch as te
except:
    te = None

try:
    import megablocks
except:
    megablocks = None

__all__ = [
    'generic_param_init_fn_',
]


def torch_default_param_init_fn_(
    module: nn.Module,
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config

    if hasattr(
        module,
        'reset_parameters',
    ) and isinstance(module.reset_parameters, Callable):
        module.reset_parameters()


def fused_init_helper_(
    module: nn.Module,
    init_fn_: Callable,
    name_param: str = 'weight',
):
    """Initializes parameters which have been fused for efficiency purposes.

    Parameter initialization is often based on the parameters shape. If a layer is fused,
    initialization should be based on the shapes of the original tensor instead of the
    shape of the fused tensor. Layers which are fused should have the _fused
    attribute. First element of _fused is the dimension along which the tensor is fused.
    Second element is a an iterable of split indices.

    Args:
        module (nn.Module): The module to initialize.
        init_fn_ (Callable): Initialization method.
        name_param (str): Name of parameter to initialize within the module.
    """
    _fused = getattr(module, '_fused', None)
    if _fused is None:
        raise RuntimeError(f'Internal logic error')

    fused_param_init_helper(getattr(module, name_param), init_fn_, _fused)


def fused_param_init_helper(
    param: torch.Tensor,
    init_fn_: Callable,
    fused_parameters: tuple[int, list[int]],
):
    """Initializes parameters that are fused together.

    Args:
        param (torch.Tensor): Tensor to initialize.
        init_fn_ (Callable): Initialization method.
        fused_parameters (tuple[int, list[int]]): First element of _fused is the dimension
            along which the tensor is fused. Second element is a an iterable of split indices.
    """
    p_ndims = param.ndim
    dim, splits = fused_parameters
    splits = (0, *splits, param.size(dim))  # type: ignore
    for s, e in zip(splits[:-1], splits[1:]):
        slice_indices = [slice(None)] * p_ndims  # type: ignore
        slice_indices[dim] = slice(s, e)
        init_fn_(param[slice_indices])  # type: ignore


def stacked_init_helper_(
    module: nn.Module,
    init_fn_: Callable,
    name_param: str = 'weight',
):
    """Initializes parameters stacked along a new dimension.

    Parameter initialization is often based on the parameters shape. If a layer is stacked,
    initialization should be based on the shapes of the original tensor instead of the
    shape of the stacked tensor. Layers which are fused should have the _stacked_dim
    attribute defining the new dimension along which they are stacked.

    Args:
        module (nn.Module): The module to initialize.
        init_fn_ (Callable): Initialization method.
        name_param (str): Name of parameter to initialize within the module.
    """
    stack_dim = getattr(module, '_stack_dim', None)
    if stack_dim is None:
        raise RuntimeError(f'Internal logic error')

    stacked_param_init_helper(getattr(module, name_param), init_fn_, stack_dim)


def stacked_param_init_helper(
    param: torch.Tensor,
    init_fn_: Callable,
    stack_dim: int,
):
    """Initialize parameters stacked along a new dimension.

    Args:
        param (torch.Tensor): Tensor to initialize.
        init_fn_ (Callable): Initialization method.
        stack_dim (int): Dimension along with parameters are stacked
    """
    p_ndims = param.ndim

    for idx in range(param.size(stack_dim)):
        slice_indices = [slice(None)] * p_ndims  # type: ignore
        slice_indices[stack_dim] = idx  # type: ignore
        init_fn_(param[slice_indices])  # type: ignore


def _flip_fan_mode(init_fn_: Callable):
    """Changes the mode of an init_fn_.

    init_fn_'s "mode" is set to operate on standard torch modules eg torch.nn.Linear.
    If a custom layer transposes its weights before they are allied such that it is
    opposite pytorch's conventions, we must flip the fan mode, from fan_in to fan_out.

    Args:
        init_fn_ (Callable): Initialization method.
    """
    _init_fn_ = deepcopy(init_fn_)
    if 'mode' in _init_fn_.keywords:
        if _init_fn_.keywords['mode'] == 'fan_in':
            _init_fn_.keywords['mode'] = 'fan_out'
        elif _init_fn_.keywords['mode'] == 'fan_out':
            _init_fn_.keywords['mode'] = 'fan_in'
    return _init_fn_


def fc_init(
    module: nn.Module,
    init_fn_: Callable,
    init_div_is_residual: Union[int, float, str, bool],
    div_is_residual: Optional[float],
    **kwargs: Any,
) -> bool:
    del kwargs  # unused, just to capture any extra args

    if isinstance(module, tuple({fcs.get(n) for n in fcs.get_all()})):
        # Linear
        if hasattr(module, '_fused'):
            fused_init_helper_(module, init_fn_)
        else:
            init_fn_(module.weight)
        if module.bias is not None:
            assert isinstance(module.bias, torch.Tensor)
            torch.nn.init.zeros_(module.bias)

        if init_div_is_residual is not False and getattr(
            module,
            '_is_residual',
            False,
        ):
            with torch.no_grad():
                module.weight.div_(div_is_residual)  # type: ignore
        return True

    return False


def embedding_init(
    module: nn.Module,
    init_fn_: Callable,
    emb_init_std: Optional[float],
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]],
    **kwargs: Any,
) -> bool:
    del kwargs  # unused, just to capture any extra args

    if isinstance(module, nn.Embedding):
        # Embedding
        if emb_init_std is not None:
            std = emb_init_std
            if std == 0:
                warnings.warn(f'Embedding layer initialized to 0.')
            emb_init_fn_ = partial(torch.nn.init.normal_, mean=0.0, std=std)
        elif emb_init_uniform_lim is not None:
            lim = emb_init_uniform_lim
            if isinstance(lim, Sequence):
                if len(lim) > 2:
                    raise ValueError(
                        f'Uniform init requires a min and a max limit. User input: {lim}.',
                    )
                if lim[0] == lim[1]:
                    warnings.warn(f'Embedding layer initialized to {lim[0]}.')
            else:
                if lim == 0:
                    warnings.warn(f'Embedding layer initialized to 0.')
                lim = [-lim, lim]
            a, b = lim
            emb_init_fn_ = partial(torch.nn.init.uniform_, a=a, b=b)
        else:
            emb_init_fn_ = init_fn_

        emb_init_fn_(module.weight)
        if module.padding_idx is not None:
            with torch.no_grad():
                module.weight[module.padding_idx].fill_(0)

        return True

    return False


def norm_init(
    module: nn.Module,
    **kwargs: Any,
) -> bool:
    del kwargs  # unused, just to capture any extra args

    if isinstance(
        module,
        tuple({norms.get(name) for name in norms.get_all()}),
    ):
        # Norm
        if hasattr(module,
                   'weight') and isinstance(module.weight, torch.Tensor):
            torch.nn.init.ones_(module.weight)
        if hasattr(module, 'bias') and isinstance(module.bias, torch.Tensor):
            torch.nn.init.zeros_(module.bias)

        return True

    return False


def multihead_attention_init(
    module: nn.Module,
    init_fn_: Callable,
    d_model: Optional[int],
    init_div_is_residual: Union[int, float, str, bool],
    div_is_residual: float,
    **kwargs: Any,
) -> bool:
    del kwargs  # unused, just to capture any extra args

    if isinstance(module, nn.MultiheadAttention):
        # torch's MultiheadAttention
        if module._qkv_same_embed_dim:
            assert module.in_proj_weight is not None
            assert module.q_proj_weight is None and module.k_proj_weight is None and module.v_proj_weight is None
            assert d_model is not None
            # in_proj_weight is actually 3 layers and should be split up for width based init
            _d = d_model
            splits = (0, _d, 2 * _d, 3 * _d)
            for s, e in zip(splits[:-1], splits[1:]):
                init_fn_(module.in_proj_weight[s:e])
        else:
            assert module.q_proj_weight is not None and module.k_proj_weight is not None and module.v_proj_weight is not None
            assert module.in_proj_weight is None
            init_fn_(module.q_proj_weight)
            init_fn_(module.k_proj_weight)
            init_fn_(module.v_proj_weight)

        # bias
        if module.in_proj_bias is not None:
            torch.nn.init.zeros_(module.in_proj_bias)
        if module.bias_k is not None:
            torch.nn.init.zeros_(module.bias_k)
        if module.bias_v is not None:
            torch.nn.init.zeros_(module.bias_v)

        # out proj
        init_fn_(module.out_proj.weight)
        if init_div_is_residual is not False and getattr(
            module.out_proj,
            '_is_residual',
            False,
        ):
            with torch.no_grad():
                module.out_proj.weight.div_(div_is_residual)
        if module.out_proj.bias is not None:
            torch.nn.init.zeros_(module.out_proj.bias)

        return True

    return False


def te_layernorm_mlp_init(
    module: nn.Module,
    init_fn_: Callable,
    **kwargs: Any,
) -> bool:
    del kwargs  # unused, just to capture any extra args

    if te is not None and isinstance(module, te.LayerNormMLP):
        if isinstance(module.layer_norm_weight, torch.Tensor):
            torch.nn.init.ones_(module.layer_norm_weight)
        if isinstance(module.layer_norm_bias, torch.Tensor):
            torch.nn.init.zeros_(module.layer_norm_bias)

        init_fn_(module.fc1_weight)
        if module.fc1_bias is not None:
            assert isinstance(module.fc1_bias, torch.Tensor)
            torch.nn.init.zeros_(module.fc1_bias)
        init_fn_(module.fc2_weight)
        if module.fc2_bias is not None:
            assert isinstance(module.fc2_bias, torch.Tensor)
            torch.nn.init.zeros_(module.fc2_bias)

        with torch.no_grad():
            module.fc2_weight.div_(div_is_residual)  # type: ignore

        return True

    return False


def moe_init(
    module: nn.Module,
    init_fn_: Callable,
    init_div_is_residual: Union[int, float, str, bool],
    div_is_residual: float,
    **kwargs: Any,
) -> bool:
    if megablocks is not None and isinstance(
        module,
        (
            megablocks.layers.moe.MoE,
            megablocks.layers.dmoe.dMoE,
            megablocks.layers.moe.ParallelMLP,
            megablocks.layers.dmoe.ParallelDroplessMLP,
        ),
    ):
        if hasattr(module, 'bias') and module.bias is not None:
            # Initialize bias to 0
            torch.nn.init.zeros_(module.bias)  # type: ignore
        return True
    elif megablocks is not None and isinstance(
        module,
        megablocks.layers.glu.SparseGLU,
    ):
        _megablocks_sparse_glu_generic_param_init_fn_(
            module,
            init_fn_,
            bool(init_div_is_residual),
            div_is_residual,
        )
        return True
    elif megablocks is not None and isinstance(
        module,
        megablocks.layers.mlp.SparseMLP,
    ):
        _megablocks_sparse_mlp_generic_param_init_fn_(
            module,
            init_fn_,
            bool(init_div_is_residual),
            div_is_residual,
        )
        return True
    elif megablocks is not None and isinstance(
        module,
        megablocks.layers.mlp.MLP,
    ):
        _megablocks_mlp_generic_param_init_fn_(
            module,
            init_fn_,
            bool(init_div_is_residual),
            div_is_residual,
        )
        return True
    elif isinstance(module, GLU):
        init_fn_(module.w1)
        init_fn_(module.v1)
        init_fn_(module.w2)
        return True
    elif isinstance(module, MLP):
        init_fn_(module.w1)
        init_fn_(module.w2)
        return True

    return False


def generic_param_init_fn_(
    module: nn.Module,
    init_fn_: Callable,
    n_layers: int,
    d_model: Optional[int] = None,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config
    # enable user to divide _is_residual weights by

    # a value which defaults to math.sqrt(2 * cfg.n_layers)
    init_div_is_residual = init_div_is_residual

    if init_div_is_residual is False:
        # not used, for pyright
        div_is_residual = 1.0
    elif init_div_is_residual is True:
        div_is_residual = math.sqrt(2 * n_layers)
    elif isinstance(init_div_is_residual,
                    float) or isinstance(init_div_is_residual, int):
        div_is_residual = init_div_is_residual
    elif init_div_is_residual.isnumeric():
        # do not trust YAML parsing to always convert numbers to numbers
        div_is_residual = float(init_div_is_residual)
    else:
        # not used, for pyright
        div_is_residual = 1.0
        raise ValueError(
            f'Expected init_div_is_residual to be boolean or numeric, got {init_div_is_residual}',
        )

    all_module_init_fns = [
        module_init_fns.get(name) for name in module_init_fns.get_all()
    ]
    did_init = False
    for module_init_fn in all_module_init_fns:
        did_init = module_init_fn(
            module=module,
            init_fn_=init_fn_,
            d_model=d_model,
            init_div_is_residual=init_div_is_residual,
            div_is_residual=div_is_residual,
            emb_init_std=emb_init_std,
            emb_init_uniform_lim=emb_init_uniform_lim,
        )

        if did_init:
            break

    if not did_init:
        for _ in module.parameters(recurse=False):
            # raise error if uninitialized module has any parameters
            raise NotImplementedError(
                f'{module.__class__.__name__} parameters are not initialized by any of the registered module_init_fns. '
                +
                'Please add an appropriate module_init_fn to the registry. Currently registered module_init_fns are: '
                + ', '.join(module_init_fns.get_all()),
            )


def _megablocks_sparse_mlp_generic_param_init_fn_(
    module: nn.Module,
    init_fn_: Callable,
    init_div_is_residual: bool = False,
    div_is_residual: float = 1.0,
):
    """Initializes MegaBlocks MLP.

    To enable elastic deterministic initialization, this method creates the entire
    weight matrix then slice into the weight tensors such that the sampled weights
    should not vary between moe world size for the same random seed.

    Args:
        module (nn.Module): The module to initialize.
        init_fn_ (Callable): Initialization method.
        init_div_is_residual (bool): Flag enabling parameters tagged with _is_residual
            flag to be divided by div_is_residual.
        div_is_residual (float): The value by which parameter initialization is divided
            if init_div_is_residual flag is enabled.
    """
    expert_process_group_size, rank = 1, 0
    if module.expert_parallel_group is not None:
        expert_process_group_size = int(
            module.expert_parallel_group.size(),
        )  # type: ignore
        rank = int(module.expert_parallel_group.rank())  # type: ignore

    hidden_size = int(module.hidden_size)  # type: ignore

    # Initialize w1
    w1 = module.w1
    if isinstance(w1, DTensor):
        w1 = w1._local_tensor
    w1_size = list(w1.shape)  # type: ignore
    w1_size[0] = w1_size[0] * expert_process_group_size

    n_exp = w1_size[0] // hidden_size
    _fused = (0, [(n + 1) * hidden_size for n in range(n_exp - 1)])

    _w1 = w1.new_empty(w1_size)  # type: ignore
    fused_param_init_helper(_w1, init_fn_, _fused)
    _w1_local = _w1.chunk(expert_process_group_size, dim=0)[rank]
    with torch.no_grad():
        w1.copy_(_w1_local)  # type: ignore

    # Initialize w2
    w2 = module.w2
    if isinstance(w2, DTensor):
        w2 = w2._local_tensor
    w2_size = list(w2.shape)  # type: ignore
    w2_size[0] = w2_size[0] * expert_process_group_size
    _w2 = w2.new_empty(w2_size)  # type: ignore
    # MegaBlocks operates on w2 as x @ w2, so needs flipped fan mode
    fused_param_init_helper(_w2, _flip_fan_mode(init_fn_), _fused)
    _w2_local = _w2.chunk(expert_process_group_size, dim=0)[rank]
    with torch.no_grad():
        w2.copy_(_w2_local)  # type: ignore
    if init_div_is_residual is not False:
        with torch.no_grad():
            w2.div_(div_is_residual)  # type: ignore


def _megablocks_sparse_glu_generic_param_init_fn_(
    module: nn.Module,
    init_fn_: Callable,
    init_div_is_residual: bool = False,
    div_is_residual: float = 1.0,
):
    """Initializes MegaBlocks Sparse GLU.

    Extends the Megablocks Sparse MLP case to an additional weight v1 for GLUs.
    This additional weight v1 has the same initialization procedure as w1 for MLPs.

    Args:
        module (nn.Module): The module to initialize.
        init_fn_ (Callable): Initialization method.
        init_div_is_residual (bool): Flag enabling parameters tagged with _is_residual
            flag to be divided by div_is_residual.
        div_is_residual (float): The value by which parameter initialization is divided
            if init_div_is_residual flag is enabled.
    """
    # Init for w1 and w2 matrices
    _megablocks_sparse_mlp_generic_param_init_fn_(
        module=module,
        init_fn_=init_fn_,
        init_div_is_residual=init_div_is_residual,
        div_is_residual=div_is_residual,
    )

    # Init ported from _megablocks_sparse_mlp_generic_param_init_fn_ for v1
    expert_process_group_size, rank = 1, 0
    if module.expert_parallel_group is not None:
        expert_process_group_size = int(
            module.expert_parallel_group.size(),
        )  # type: ignore
        rank = int(module.expert_parallel_group.rank())  # type: ignore

    hidden_size = int(module.hidden_size)  # type: ignore

    # Separately initialize v1
    v1 = module.v1
    if isinstance(v1, DTensor):
        v1 = v1._local_tensor
    v1_size = list(v1.shape)  # type: ignore
    v1_size[0] = v1_size[0] * expert_process_group_size

    n_exp = v1_size[0] // hidden_size
    _fused = (0, [(n + 1) * hidden_size for n in range(n_exp - 1)])

    _v1 = v1.new_empty(v1_size)  # type: ignore
    fused_param_init_helper(_v1, init_fn_, _fused)
    _v1_local = _v1.chunk(expert_process_group_size, dim=0)[rank]
    with torch.no_grad():
        v1.copy_(_v1_local)  # type: ignore


def _megablocks_mlp_generic_param_init_fn_(
    module: nn.Module,
    init_fn_: Callable,
    init_div_is_residual: bool = False,
    div_is_residual: float = 1.0,
):
    """Initializes MegaBlocks' MLP.

    To enable elastic deterministic initialization, this method creates the entire
    weight matrix then slice into the weight tensors such that the sampled weights
    should not vary between moe world size for the same random seed.

    Args:
        module (nn.Module): The module to initialize.
        init_fn_ (Callable): Initialization method.
        init_div_is_residual (bool): Flag enabling parameters tagged with _is_residual
            flag to be divided by div_is_residual.
        div_is_residual (float): The value by which parameter initialization is divided
            if init_div_is_residual flag is enabled.
    """
    expert_process_group_size, rank = 1, 0
    if module.expert_parallel_group is not None:
        expert_process_group_size = int(
            module.expert_parallel_group.size(),
        )  # type: ignore
        rank = int(module.expert_parallel_group.rank())  # type: ignore

    _init_fn_ = _flip_fan_mode(init_fn_)

    # Initialize w1
    w1_size = list(module.w1.shape)  # type: ignore
    w1_size[0] = w1_size[0] * expert_process_group_size
    _w1 = module.w1.new_empty(w1_size)  # type: ignore
    stacked_param_init_helper(_w1, _init_fn_, module._stack_dim)  # type: ignore
    _w1_local = _w1.chunk(expert_process_group_size, dim=0)[rank]
    with torch.no_grad():
        module.w1.copy_(_w1_local)  # type: ignore

    # Initialize w2
    w2_size = list(module.w2.shape)  # type: ignore
    w2_size[0] = w2_size[0] * expert_process_group_size
    _w2 = module.w2.new_empty(w2_size)  # type: ignore
    stacked_param_init_helper(_w2, _init_fn_, module._stack_dim)  # type: ignore
    _w2_local = _w2.chunk(expert_process_group_size, dim=0)[rank]
    with torch.no_grad():
        module.w2.copy_(_w2_local)  # type: ignore
    if init_div_is_residual is not False:
        with torch.no_grad():
            module.w2.div_(div_is_residual)  # type: ignore


def _normal_init_(std: float, mean: float = 0.0):
    return partial(torch.nn.init.normal_, mean=mean, std=std)


def _normal_param_init_fn_(
    module: nn.Module,
    std: float,
    n_layers: int,
    d_model: Optional[int] = None,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config
    init_fn_ = _normal_init_(std=std)

    generic_param_init_fn_(
        module=module,
        init_fn_=init_fn_,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=init_div_is_residual,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


def baseline_param_init_fn_(
    module: nn.Module,
    init_std: Optional[float],
    n_layers: int,
    d_model: Optional[int] = None,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config
    if init_std is None:
        raise ValueError(
            "You must set model.init_config['init_std'] to a float value to use the default initialization scheme.",
        )
    _normal_param_init_fn_(
        module=module,
        std=init_std,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=init_div_is_residual,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


def small_param_init_fn_(
    module: nn.Module,
    n_layers: int,
    d_model: int,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config
    # Very close to kaiming normal
    # From Transformers without Tears (2019) - Nguyen & Salazar
    std = math.sqrt(2 / (5 * d_model))
    _normal_param_init_fn_(
        module=module,
        std=std,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=init_div_is_residual,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


def neox_param_init_fn_(
    module: nn.Module,
    n_layers: int,
    d_model: int,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    **kwargs: Any,
) -> None:
    """From section 2.3.1 of GPT-NeoX-20B:

    An Open-Source AutoregressiveLanguage Model — Black et. al. (2022)
    see https://github.com/EleutherAI/gpt-neox/blob/9610391ab319403cef079b438edd016a2443af54/megatron/model/init_functions.py#L151
    and https://github.com/EleutherAI/gpt-neox/blob/main/megatron/model/transformer.py
    """
    del kwargs  # unused, just to capture any extra args from the config
    residual_div = n_layers / math.sqrt(10)  # small std / wang std

    small_param_init_fn_(
        module=module,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=residual_div,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


def kaiming_uniform_param_init_fn_(
    module: nn.Module,
    n_layers: int,
    d_model: Optional[int] = None,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    init_gain: float = 0,
    fan_mode: str = 'fan_in',
    init_nonlinearity: str = 'leaky_relu',
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config

    kaiming_uniform_ = partial(
        nn.init.kaiming_uniform_,
        a=init_gain,
        mode=fan_mode,
        nonlinearity=init_nonlinearity,
    )

    generic_param_init_fn_(
        module=module,
        init_fn_=kaiming_uniform_,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=init_div_is_residual,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


def kaiming_normal_param_init_fn_(
    module: nn.Module,
    n_layers: int,
    d_model: Optional[int] = None,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    init_gain: float = 0,
    fan_mode: str = 'fan_in',
    init_nonlinearity: str = 'leaky_relu',
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config

    kaiming_normal_ = partial(
        torch.nn.init.kaiming_normal_,
        a=init_gain,
        mode=fan_mode,
        nonlinearity=init_nonlinearity,
    )

    generic_param_init_fn_(
        module=module,
        init_fn_=kaiming_normal_,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=init_div_is_residual,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


def xavier_uniform_param_init_fn_(
    module: nn.Module,
    n_layers: int,
    d_model: Optional[int] = None,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    init_gain: float = 0,
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config
    xavier_uniform_ = partial(torch.nn.init.xavier_uniform_, gain=init_gain)

    generic_param_init_fn_(
        module=module,
        init_fn_=xavier_uniform_,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=init_div_is_residual,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


def xavier_normal_param_init_fn_(
    module: nn.Module,
    n_layers: int,
    d_model: Optional[int] = None,
    init_div_is_residual: Union[int, float, str, bool] = True,
    emb_init_std: Optional[float] = None,
    emb_init_uniform_lim: Optional[Union[tuple[float, float], float]] = None,
    init_gain: float = 0,
    **kwargs: Any,
) -> None:
    del kwargs  # unused, just to capture any extra args from the config
    xavier_normal_ = partial(torch.nn.init.xavier_normal_, gain=init_gain)

    generic_param_init_fn_(
        module=module,
        init_fn_=xavier_normal_,
        d_model=d_model,
        n_layers=n_layers,
        init_div_is_residual=init_div_is_residual,
        emb_init_std=emb_init_std,
        emb_init_uniform_lim=emb_init_uniform_lim,
    )


param_init_fns.register('default_', func=torch_default_param_init_fn_)
param_init_fns.register('baseline_', func=baseline_param_init_fn_)
param_init_fns.register('kaiming_uniform_', func=kaiming_uniform_param_init_fn_)
param_init_fns.register('kaiming_normal_', func=kaiming_normal_param_init_fn_)
param_init_fns.register('neox_init_', func=neox_param_init_fn_)
param_init_fns.register('small_init_', func=small_param_init_fn_)
param_init_fns.register('xavier_uniform_', func=xavier_uniform_param_init_fn_)
param_init_fns.register('xavier_normal_', func=xavier_normal_param_init_fn_)

module_init_fns.register('fc', func=fc_init)
module_init_fns.register('embedding', func=embedding_init)
module_init_fns.register('norm', func=norm_init)
module_init_fns.register('multihead_attention', func=multihead_attention_init)
module_init_fns.register('te_layernorm_mlp', func=te_layernorm_mlp_init)
module_init_fns.register('moe', func=moe_init)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""GPT Blocks used for the GPT Model."""

import copy
from typing import Any, Optional

import torch
import torch.nn as nn

from llmfoundry.layers_registry import ffns_with_norm
from llmfoundry.models.layers.layer_builders import (
    build_attention_layer,
    build_ffn,
    build_norm,
)
from llmfoundry.models.utils.config_defaults import (
    attn_config_defaults,
    fc_type_defaults,
)

try:
    from flash_attn.bert_padding import unpad_input, pad_input  # type: ignore # yapf: disable # isort: skip
except:
    unpad_input, pad_input = None, None

__all__ = [
    'MPTBlock',
    'FusedNormAttentionNorm',
]


class MPTBlock(nn.Module):

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        expansion_ratio: int,
        attn_config: Optional[dict] = None,
        ffn_config: Optional[dict] = None,
        resid_pdrop: float = 0.0,
        norm_type: str = 'low_precision_layernorm',
        norm_eps: float = 1e-05,
        fc_type: Optional[dict[str, Any]] = None,
        device: Optional[str] = None,
        no_bias: bool = False,
        use_pad_tok_in_ffn: bool = True,
        **kwargs: Any,
    ):
        if attn_config is None:
            attn_config = attn_config_defaults

        if ffn_config is None:
            self.ffn_config: dict[str, Any] = {
                'ffn_type': 'mptmlp',
            }
        else:
            self.ffn_config = ffn_config

        if fc_type is None:
            fc_type = copy.deepcopy(fc_type_defaults)
        fc_type['bias'] = not no_bias
        fc_type['device'] = device

        self.ffn_config['fc_type'] = fc_type

        self.fuse_norm_attn_norm = kwargs.get('fuse_norm_attn_norm', False)

        del kwargs  # unused, just to capture any extra args from the config
        super().__init__()

        ffn_type = self.ffn_config['ffn_type']
        ffn_has_norm = ffn_type in ffns_with_norm

        if self.fuse_norm_attn_norm:
            self.norm_attn_norm = FusedNormAttentionNorm(
                d_model=d_model,
                n_heads=n_heads,
                args_to_exclude_in_attn_class=self.
                args_to_exclude_in_attn_class,
                attn_config=attn_config,
                ffn_has_norm=ffn_has_norm,
                fc_type=fc_type,
                resid_pdrop=resid_pdrop,
                norm_type=norm_type,
                norm_eps=norm_eps,
                device=device,
                no_bias=no_bias,
            )
        else:
            assert isinstance(attn_config['attn_type'], str)
            # Necessary to avoid passing extraneous args into attn_class while allowing the use of **kwargs
            attn_config_subset_for_attn_class = {
                k: v
                for k, v in attn_config.items()
                if k not in self.args_to_exclude_in_attn_class
            }

            self.norm_1 = build_norm(
                name=norm_type.lower(),
                normalized_shape=d_model,
                eps=norm_eps,
                device=device,
            )
            self.attn = build_attention_layer(
                name=attn_config['attn_type'],
                attn_kwargs={
                    'd_model': d_model,
                    'n_heads': n_heads,
                    'fc_type': fc_type,
                    'device': device,
                    'bias': not no_bias,
                    **attn_config_subset_for_attn_class,
                },
            )
            self.norm_2 = None
            if not ffn_has_norm:
                self.norm_2 = build_norm(
                    name=norm_type.lower(),
                    normalized_shape=d_model,
                    eps=norm_eps,
                    device=device,
                )

        self.ffn = build_ffn(
            name=ffn_type,
            d_model=d_model,
            expansion_ratio=expansion_ratio,
            device=device,
            bias=not no_bias,
            ffn_kwargs=self.ffn_config,
        )

        self.resid_attn_dropout = nn.Dropout(resid_pdrop)
        self.resid_ffn_dropout = nn.Dropout(resid_pdrop)
        self.use_pad_tok_in_ffn = use_pad_tok_in_ffn

    @property
    def args_to_exclude_in_attn_class(self):
        return {
            'attn_type',
            'alibi',
            'attn_uses_sequence_id',
            'alibi_bias_max',
            'rope',
            'rope_theta',
            'rope_impl',
            'rope_dail_config',
            'rope_hf_config',
        }

    def forward(
        self,
        x: torch.Tensor,
        past_key_value: Optional[tuple[torch.Tensor, torch.Tensor]] = None,
        attn_bias: Optional[torch.Tensor] = None,
        rotary_emb_w_meta_info: Optional[dict] = None,
        attention_mask: Optional[torch.ByteTensor] = None,
        is_causal: bool = True,
        output_attentions: bool = False,
        alibi_slopes: Optional[torch.Tensor] = None,
        flash_attn_padding_info: Optional[dict[str, torch.Tensor]] = None,
        prev_layer_key_value: Optional[tuple[torch.Tensor,
                                             torch.Tensor]] = None,
        key_value_states: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, Optional[torch.Tensor], Optional[tuple[
        torch.Tensor, torch.Tensor]]]:
        extra_kwargs = {}
        if prev_layer_key_value is not None:
            extra_kwargs['prev_layer_key_value'] = prev_layer_key_value
        if key_value_states is not None:
            extra_kwargs['key_value_states'] = key_value_states

        if self.fuse_norm_attn_norm:
            x, m, attn_weights, past_key_value = self.norm_attn_norm(
                x,
                past_key_value=past_key_value,
                attn_bias=attn_bias,
                rotary_emb_w_meta_info=rotary_emb_w_meta_info,
                attention_mask=attention_mask,
                is_causal=is_causal,
                output_attentions=output_attentions,
                alibi_slopes=alibi_slopes,
                flash_attn_padding_info=flash_attn_padding_info,
                **extra_kwargs,
            )
        else:
            a = self.norm_1(x)
            b, attn_weights, past_key_value = self.attn(
                a,
                past_key_value=past_key_value,
                attn_bias=attn_bias,
                rotary_emb_w_meta_info=rotary_emb_w_meta_info,
                attention_mask=attention_mask,
                is_causal=is_causal,
                needs_weights=output_attentions,
                alibi_slopes=alibi_slopes,
                flash_attn_padding_info=flash_attn_padding_info,
                **extra_kwargs,
            )
            x = x + self.resid_attn_dropout(b)
            m = x
            if self.norm_2 is not None:
                m = self.norm_2(x)

        n = self.apply_ffn(attention_mask, m)
        # In the following line we move the `x` tensor to the same devices as the output of ffn layer. This operation should be a no-op during training.
        # This is done to fix pipeline parallel generation using hf.generate. Please see this comment for details: https://github.com/mosaicml/llm-foundry/pull/1332#issue-2386827204
        x = x.to(device=n.device,
                ) + self.resid_ffn_dropout(n).to(device=n.device,)
        return x, attn_weights, past_key_value

    def apply_ffn(
        self,
        attention_mask: Optional[torch.ByteTensor],
        m: torch.Tensor,
    ) -> torch.Tensor:
        """Apply feed forward layers to the input.

        Args:
            attention_mask (Optional[torch.ByteTensor]): The attention mask.
            m (torch.Tensor): The input.

        Returns:
            n (torch.Tensor): The output.
        """
        batch_size, seq_len = m.size()[:2]
        indices = None
        if not self.use_pad_tok_in_ffn and attention_mask is not None:
            assert unpad_input is not None
            attention_mask = self.slice_attention_mask(attention_mask, seq_len)
            m, indices, _, _ = unpad_input(m, attention_mask)
        n = self.ffn(m)
        if not self.use_pad_tok_in_ffn and attention_mask is not None:
            assert pad_input is not None
            n = pad_input(n, indices, batch_size, seq_len)
        return n

    def slice_attention_mask(
        self,
        attention_mask: torch.ByteTensor,
        seq_len: int,
    ) -> torch.ByteTensor:
        """Slice attention mask to the correct size.

        Can be overridden by subclasses to apply different slicing logic.

        Args:
            attention_mask (torch.ByteTensor): The attention mask.
            seq_len (int): The sequence length.

        Returns:
            torch.ByteTensor: The sliced attention mask.
        """
        return attention_mask


class FusedNormAttentionNorm(nn.Module):

    def __init__(
        self,
        d_model: int,
        n_heads: int,
        args_to_exclude_in_attn_class: set[str],
        attn_config: Optional[dict] = None,
        ffn_has_norm: bool = False,
        fc_type: Optional[dict[str, Any]] = None,
        resid_pdrop: float = 0.0,
        norm_type: str = 'low_precision_layernorm',
        norm_eps: float = 1e-05,
        device: Optional[str] = None,
        no_bias: bool = False,
        **kwargs: Any,
    ):
        super().__init__()
        assert attn_config is not None
        assert isinstance(attn_config['attn_type'], str)

        # Usually, fc_type dict should be passed in through MPTBlock's __init__ function.
        if fc_type is None:
            fc_type = copy.deepcopy(fc_type_defaults)
            fc_type['bias'] = not no_bias
            fc_type['device'] = device

        # Necessary to avoid passing extraneous args into attn_class while allowing the use of **kwargs
        attn_config_subset_for_attn_class = {
            k: v
            for k, v in attn_config.items()
            if k not in args_to_exclude_in_attn_class
        }
        self.norm_1 = build_norm(
            name=norm_type.lower(),
            normalized_shape=d_model,
            eps=norm_eps,
            device=device,
        )
        self.attn = build_attention_layer(
            name=attn_config['attn_type'],
            attn_kwargs={
                'd_model': d_model,
                'n_heads': n_heads,
                'fc_type': fc_type,
                'device': device,
                'bias': not no_bias,
                **attn_config_subset_for_attn_class,
            },
        )

        self.norm_2 = None
        if not ffn_has_norm:
            self.norm_2 = build_norm(
                name=norm_type.lower(),
                normalized_shape=d_model,
                eps=norm_eps,
                device=device,
            )
        self.resid_attn_dropout = nn.Dropout(resid_pdrop)

    def forward(
        self,
        x: torch.Tensor,
        past_key_value: Optional[tuple[torch.Tensor, torch.Tensor]] = None,
        attn_bias: Optional[torch.Tensor] = None,
        rotary_emb_w_meta_info: Optional[dict] = None,
        attention_mask: Optional[torch.ByteTensor] = None,
        is_causal: bool = True,
        output_attentions: bool = False,
        alibi_slopes: Optional[torch.Tensor] = None,
        flash_attn_padding_info: Optional[dict[str, torch.Tensor]] = None,
        prev_layer_key_value: Optional[tuple[torch.Tensor,
                                             torch.Tensor]] = None,
        key_value_states: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor],
               Optional[tuple[torch.Tensor, torch.Tensor]]]:
        a = self.norm_1(x)
        extra_kwargs = {}
        if prev_layer_key_value is not None:
            extra_kwargs['prev_layer_key_value'] = prev_layer_key_value
        if key_value_states is not None:
            extra_kwargs['key_value_states'] = key_value_states

        b, attn_weights, past_key_value = self.attn(
            a,
            past_key_value=past_key_value,
            attn_bias=attn_bias,
            rotary_emb_w_meta_info=rotary_emb_w_meta_info,
            attention_mask=attention_mask,
            is_causal=is_causal,
            needs_weights=output_attentions,
            alibi_slopes=alibi_slopes,
            flash_attn_padding_info=flash_attn_padding_info,
            **extra_kwargs,
        )
        x = x + self.resid_attn_dropout(b)
        m = x
        if self.norm_2 is not None:
            m = self.norm_2(x)

        return x, m, attn_weights, past_key_value



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.models.mpt.configuration_mpt import MPTConfig
from llmfoundry.models.mpt.modeling_mpt import (
    ComposerMPTCausalLM,
    MPTForCausalLM,
    MPTModel,
    MPTPreTrainedModel,
)

__all__ = [
    'MPTPreTrainedModel',
    'MPTModel',
    'MPTForCausalLM',
    'ComposerMPTCausalLM',
    'MPTConfig',
]
